# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing different diagnostics for mesoscales.

The diagnostics here are applicable at mesoscales and apply generally
rather than for specific aspects of mesoscale meteorology (e.g. convection).
For specific aspects, the user is referred to other modules available in
CSET.

"""

import logging

import iris
from scipy.ndimage import gaussian_filter, uniform_filter

from CSET.operators._utils import get_cube_yxcoordname


def spatial_perturbation_field(
    original_field: iris.cube.Cube,
    apply_gaussian_filter: bool = True,
    filter_scale: int = 40,
) -> iris.cube.Cube:
    """Calculate a spatial perturbation field.

    Parameters
    ----------
    original_field: iris.cube.Cube
        Iris cube containing data to smooth, supporting multiple dimensions
        (at least two spatial dimensions must be supplied, i.e. 2D).
    apply_gaussian_filter: boolean, optional
        If set to True a Gaussian filter is applied; if set to False
        a Uniform filter is applied.
        Default is True.
    filter_scale: int, optional
        Scale at which to define the filter in grid boxes. If the
        filter is a Gaussian convolution this value represents the
        standard deviation of the Gaussian kernel.
        Default is 40 grid boxes.

    Returns
    -------
    pert_field: iris.cube.Cube
        An iris cube of the spatial perturbation field.

    Notes
    -----
    In mesoscale meteorology the perturbation field is more important than the
    balanced flows for process understanding. This function is designed
    to create spatial perturbation fields based on smoothing with a Gaussian
    kernel or a uniform kernel.

    The kernels are defined by the filter_scale, which for mesoscale
    perturbations should be between an approximate cloud separation distance,
    of 30 km, and synoptic scale variations (1000 km). In practice any
    value between these ranges should provided broadly consistent results (e.g.
    [Flacketal2016]_). The Gaussian kernel will give greater importance to
    areas closer to the event and will produce a smooth perturbation field.
    The uniform kernel will produce a smooth perturbation field but will not
    give local features as much prominence.

    Caution should be applied to boundaries, particularly if the domain is of
    variable resolution, as some numerical artifacts could be introduced.

    References
    ----------
    .. [Flacketal2016] Flack, D.L.A., Plant, R.S., Gray, S.L., Lean, H.W.,
       Keil, C. and Craig, G.C. (2016) "Characterisation of Convective
       Regimes over the British Isles." Quarterly Journal of the Royal
       Meteorological Society, vol. 142, 1541-1553. doi:10.1002/qj.2758

    Examples
    --------
    >>> Temperature_perturbation = meso.spatial_perturbation_fields(Temp,
                                   gaussian_filter=True,filter_scale=40)
    >>> iplt.pcolormesh(Temperature_perturabtion[0,:,:],cmap=mpl.cm.bwr)
    >>> plt.gca().coastlines('10m')
    >>> plt.clim(-5,5)
    >>> plt.colorbar()
    >>> plt.show()

    """
    pert_field = original_field.copy()
    # find axes of spatial coordinates in field
    coords = [coord.name() for coord in original_field.coords()]
    # axes tuple containing latitude, longitude coordinate name.
    axes = (
        coords.index(get_cube_yxcoordname(original_field)[0]),
        coords.index(get_cube_yxcoordname(original_field)[1]),
    )
    # apply convolution depending on type used
    if apply_gaussian_filter:
        filter_type = "Gaussian"
        logging.info("Gaussian filter applied.")
        pert_field.data -= gaussian_filter(original_field.data, filter_scale, axes=axes)
    else:
        logging.info("Uniform filter applied.")
        filter_type = "Uniform"
        pert_field.data -= uniform_filter(original_field.data, filter_scale, axes=axes)
    # provide attributes to cube to indicate spatial perturbation field
    pert_field.attributes["perturbation_field"] = (
        f"{filter_type}_with_{str(filter_scale)}_grid_point_filter_scale"
    )
    return pert_field



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for run_cset_recipe workflow utility."""

import os
import subprocess
import zipfile
from pathlib import Path

import pytest

from CSET._workflow_utils import run_cset_recipe


def test_subprocess_env():
    """Test subprocess_env function."""
    expected = dict(os.environ)
    actual = run_cset_recipe.subprocess_env()
    assert actual == expected


def test_recipe_file(monkeypatch, tmp_working_dir):
    """Bundled recipe file is written to disk."""
    monkeypatch.setenv("CSET_RECIPE_NAME", "CAPE_ratio_plot.yaml")
    recipe_path = run_cset_recipe.recipe_file()
    assert recipe_path == "CAPE_ratio_plot.yaml"
    assert Path(recipe_path).is_file()


def test_recipe_id(monkeypatch, tmp_working_dir):
    """Extract recipe ID from recipe file."""

    def mock_recipe_file():
        with open("recipe.yaml", "wt", encoding="UTF-8") as fp:
            fp.write("title: Recipe Title\nsteps: [{operator: misc.noop}]")
        return "recipe.yaml"

    monkeypatch.setenv("MODEL_IDENTIFIERS", "1")
    monkeypatch.setattr(run_cset_recipe, "recipe_file", mock_recipe_file)
    expected = "m1_recipe_title"
    actual = run_cset_recipe.recipe_id()
    assert actual == expected


def test_recipe_id_invalid_recipe(monkeypatch, tmp_working_dir):
    """Invalid recipe file raises exception."""

    def mock_recipe_file():
        with open("recipe.yaml", "wt", encoding="UTF-8") as fp:
            fp.write("Not a recipe!")
        return "recipe.yaml"

    monkeypatch.setenv("MODEL_IDENTIFIERS", "1")
    monkeypatch.setattr(run_cset_recipe, "recipe_file", mock_recipe_file)
    with pytest.raises(subprocess.CalledProcessError):
        run_cset_recipe.recipe_id()


def test_output_directory(monkeypatch):
    """Output directory correctly interpreted."""

    def mock_recipe_id():
        return "recipe_id"

    monkeypatch.setattr(run_cset_recipe, "recipe_id", mock_recipe_id)
    monkeypatch.setenv("CYLC_WORKFLOW_SHARE_DIR", "/share")
    monkeypatch.setenv("CYLC_TASK_CYCLE_POINT", "20000101T0000Z")
    actual = run_cset_recipe.output_directory()
    expected = "/share/web/plots/recipe_id_20000101T0000Z"
    assert actual == expected


def test_data_directories(monkeypatch):
    """Data directory correctly interpreted."""
    monkeypatch.setenv("ROSE_DATAC", "/share/cycle/20000101T0000Z")
    monkeypatch.setenv("MODEL_IDENTIFIERS", "1")
    expected = ["/share/cycle/20000101T0000Z/data/1"]
    actual = run_cset_recipe.data_directories()
    assert actual == expected


def test_data_directories_multiple_cases(monkeypatch):
    """Data directory correctly interpreted for multiple cases."""
    monkeypatch.setenv("CYLC_WORKFLOW_SHARE_DIR", "/share")
    monkeypatch.setenv("DO_CASE_AGGREGATION", "True")
    monkeypatch.setenv("MODEL_IDENTIFIERS", "1")
    expected = ["/share/cycle/*/data/1"]
    actual = run_cset_recipe.data_directories()
    assert actual == expected


def test_create_diagnostic_archive(tmp_path):
    """Create ZIP archive of output."""
    # Create dummy output files.
    files = {"one", "two", "three"}
    for filename in files:
        (tmp_path / filename).touch()

    run_cset_recipe.create_diagnostic_archive(tmp_path)
    archive_path = tmp_path / "diagnostic.zip"
    assert archive_path.is_file()
    with zipfile.ZipFile(archive_path, "r") as archive:
        # Check all files are now in archive.
        assert set(archive.namelist()) == files


def test_entrypoint(monkeypatch):
    """Check that run_cset_recipe.run() calls the correct function."""
    function_ran = False

    def assert_true():
        nonlocal function_ran
        function_ran = True

    monkeypatch.setattr(run_cset_recipe, "run_recipe_steps", assert_true)
    run_cset_recipe.run()
    assert function_ran, "Function did not run!"


def test_run_recipe_steps(monkeypatch, tmp_working_dir):
    """Test run recipe steps correctly runs CSET and creates an archive."""

    def mock_func(*args, **kwargs):
        return ""

    def mock_data_dirs(*args, **kwargs):
        return [""]

    monkeypatch.setattr(subprocess, "run", mock_func)
    monkeypatch.setattr(run_cset_recipe, "create_diagnostic_archive", mock_func)
    monkeypatch.setattr(run_cset_recipe, "recipe_file", mock_func)
    monkeypatch.setattr(run_cset_recipe, "output_directory", mock_func)
    monkeypatch.setattr(run_cset_recipe, "data_directories", mock_data_dirs)
    run_cset_recipe.run_recipe_steps()


def test_run_recipe_steps_exception(monkeypatch, tmp_working_dir):
    """Test run recipe steps correctly raises exception on cset bake error."""

    def mock_subprocess_run(*args, **kwargs):
        raise subprocess.CalledProcessError(1, args, b"", b"")

    def mock_func(*args, **kwargs):
        return ""

    def mock_data_dirs(*args, **kwargs):
        return [""]

    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    monkeypatch.setattr(run_cset_recipe, "create_diagnostic_archive", mock_func)
    monkeypatch.setattr(run_cset_recipe, "recipe_file", mock_func)
    monkeypatch.setattr(run_cset_recipe, "output_directory", mock_func)
    monkeypatch.setattr(run_cset_recipe, "data_directories", mock_data_dirs)
    with pytest.raises(subprocess.CalledProcessError):
        run_cset_recipe.run_recipe_steps()



#! /usr/bin/env python3

"""Send a notification email linking to the output page.

Sends a notification email to the user of the workflow when it has completed,
with a link to the output webpage.
"""

import os
import subprocess


def get_home_page_address():
    """Derive the address of the output page based directory structure."""
    try:
        web_dir = os.environ["WEB_DIR"]
    except KeyError as err:
        raise ValueError("WEB_DIR not set.") from err
    try:
        web_addr = os.environ["WEB_ADDR"]
    except KeyError as err:
        raise ValueError("WEB_ADDR not set.") from err

    path = web_dir.split("/public_html/")
    if len(path) == 1:
        path = os.getenv("WEB_DIR").split("/Public/")
    if len(path) == 1:
        raise ValueError("Cannot determine web address.")
    base_address = web_addr.strip("/")
    return f"{base_address}/{path[-1]}"


def run():
    """Run workflow script."""
    subject = "CSET webpage ready"
    try:
        body = f"The webpage for your run of CSET is now ready. You can view it here:\n{get_home_page_address()}"
    except ValueError:
        body = "The webpage for your run of CSET is now ready, though the address could not be determined.\nCheck that WEB_ADDR and WEB_DIR are set correctly, then consider filing a bug report at https://github.com/MetOffice/CSET"
    subprocess.run(
        f'printf "{body}" | mail -s "{subject}" -S "from=notifications" "$USER"',
        check=True,
        shell=True,
    )



#! /usr/bin/env python3
# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sort a JSON file."""

import argparse
import hashlib
import json
import os
import sys


def sort_dict(d: dict) -> dict:
    """Recursively sort a dictionary."""
    # Thank you to https://stackoverflow.com/a/47882384
    return {k: sort_dict(v) if isinstance(v, dict) else v for k, v in sorted(d.items())}


def main():
    """Sort the keys in a JSON file, overwriting the original file."""
    parser = argparse.ArgumentParser(description="Sort a JSON file alphabetically.")
    parser.add_argument("input_file", type=str, help="JSON file to sort.")
    args = parser.parse_args()

    try:
        with open(args.input_file, "rt", encoding="UTF-8") as fp:
            data = json.load(fp)
    except FileNotFoundError:
        print(f"Input file {args.input_file} not found", file=sys.stderr)
        return 1
    except json.JSONDecodeError:
        print(f"Input file {args.input_file} is not valid JSON", file=sys.stderr)
        return 1

    if isinstance(data, dict):
        data = sort_dict(data)

    output_file = f".tmp-{args.input_file}-sorted.json".replace("/", "")
    with open(output_file, "wt", encoding="UTF-8") as fp:
        json.dump(data, fp, indent=2)
        # End file with a newline.
        fp.write("\n")

    # Check if sorting changed the file.
    with open(args.input_file, "rb") as fp:
        original_digest = hashlib.file_digest(fp, "sha256").digest()
    with open(output_file, "rb") as fp:
        new_digest = hashlib.file_digest(fp, "sha256").digest()

    if original_digest == new_digest:
        # File unchanged, don't overwrite and return zero.
        os.remove(output_file)
        return 0
    else:
        # File changed, overwrite and return non-zero.
        # Atomically replace original file via rename.
        os.rename(output_file, args.input_file)
        return 1


if __name__ == "__main__":
    sys.exit(main())



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test collapse operators."""

import iris
import iris.cube
import numpy as np
import pytest

from CSET.operators import collapse, read


@pytest.fixture()
def long_forecast() -> iris.cube.Cube:
    """Get long_forecast to run tests on."""
    return read.read_cube(
        "tests/test_data/long_forecast_air_temp_fcst_1.nc", "air_temperature"
    )


@pytest.fixture()
def long_forecast_multi_day() -> iris.cube.Cube:
    """Get long_forecast_multi_day to run tests on."""
    return read.read_cube(
        "tests/test_data/long_forecast_air_temp_multi_day.nc", "air_temperature"
    )


@pytest.fixture()
def long_forecast_many_cubes() -> iris.cube.Cube:
    """Get long_forecast_may_cubes to run tests on."""
    return read.read_cubes(
        "tests/test_data/long_forecast_air_temp_fcst_*.nc", "air_temperature"
    )


@pytest.fixture()
def medium_forecast() -> iris.cube.Cube:
    """Get medium forecast with monotonic time coordinate."""
    return iris.load_cube(
        "tests/test_data/medium_forecast_air_temp_monotonic.nc", "air_temperature"
    )


def test_collapse(cube):
    """Reduces dimension of cube."""
    # Test collapsing a single coordinate.
    collapsed_cube = collapse.collapse(cube, "time", "MEAN")
    expected_cube = (
        "<iris 'Cube' of air_temperature / (K) (grid_latitude: 17; grid_longitude: 13)>"
    )
    assert repr(collapsed_cube) == expected_cube

    # Test collapsing two coordinates.
    collapsed_cube = collapse.collapse(
        cube, ["grid_latitude", "grid_longitude"], "MEAN"
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (time: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_cubelist(cube):
    """Reduce dimensions of multiple cubes in a CubeList."""
    cubes = iris.cube.CubeList([cube, cube])
    collapsed_cubes = collapse.collapse(cubes, "time", "MEAN")
    assert len(collapsed_cubes) == 2
    expected_cube = (
        "<iris 'Cube' of air_temperature / (K) (grid_latitude: 17; grid_longitude: 13)>"
    )
    for collapsed_cube in collapsed_cubes:
        assert repr(collapsed_cube) == expected_cube


def test_collapse_percentile(cube):
    """Reduce dimension of a cube with a PERCENTILE aggregation."""
    with pytest.raises(ValueError):
        collapse.collapse(cube, "time", "PERCENTILE")

    # Test collapsing a single coordinate.
    collapsed_cube = collapse.collapse(
        cube, "time", "PERCENTILE", additional_percent=75
    )
    expected_cube = (
        "<iris 'Cube' of air_temperature / (K) (grid_latitude: 17; grid_longitude: 13)>"
    )
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_hour_of_day(long_forecast):
    """Convert and aggregates time dimension by hour of day."""
    collapsed_cube = collapse.collapse_by_hour_of_day(
        long_forecast, "MEAN", multi_case=False
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (hour: 24; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_hour_of_day_fail(long_forecast):
    """Test failing due to multi_case set to True."""
    with pytest.raises(TypeError):
        collapse.collapse_by_hour_of_day(long_forecast, "MEAN")


def test_collapse_by_hour_of_day_percentile(long_forecast):
    """Convert and aggregate time dimension by hour of day with percentiles."""
    # Test collapsing long forecast.
    collapsed_cube = collapse.collapse_by_hour_of_day(
        long_forecast, "PERCENTILE", additional_percent=[25, 75], multi_case=False
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (percentile_over_hour: 2; hour: 24; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_hour_of_day_percentile_fail(long_forecast):
    """Test failing due to non-specified additional_percent."""
    with pytest.raises(ValueError):
        collapse.collapse_by_hour_of_day(long_forecast, "PERCENTILE", multi_case=False)


def test_collapse_by_hour_of_day_multi_forecast_cube(long_forecast_multi_day):
    """Convert and aggregates time dimension by hour of day for a multi day cube."""
    collapsed_cube = collapse.collapse_by_hour_of_day(long_forecast_multi_day, "MEAN")
    expected_cube = "<iris 'Cube' of air_temperature / (K) (hour: 24; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_hour_of_day_multi_forecast_cube_fail(long_forecast_multi_day):
    """Test failing due to multi_case set to False."""
    with pytest.raises(TypeError):
        collapse.collapse_by_hour_of_day(
            long_forecast_multi_day, "MEAN", multi_case=False
        )


def test_collapse_by_hour_of_day_multi_forecast_cubelist(long_forecast_many_cubes):
    """Convert and aggregates time dimension by hour of day for a CubeList."""
    collapsed_cube = collapse.collapse_by_hour_of_day(long_forecast_many_cubes, "MEAN")
    expected_cube = "<iris 'Cube' of air_temperature / (K) (hour: 24; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_hour_of_day_multi_forecast_cubelist_fail(long_forecast_many_cubes):
    """Test failing due to multi_case set to False."""
    with pytest.raises(TypeError):
        collapse.collapse_by_hour_of_day(
            long_forecast_many_cubes, "MEAN", multi_case=False
        )


def test_collapse_by_lead_time_single_cube(long_forecast_multi_day):
    """Check cube collapse by lead time."""
    calculated_cube = collapse.collapse(
        long_forecast_multi_day, "forecast_reference_time", "MEAN"
    )
    assert np.allclose(
        calculated_cube.data,
        collapse.collapse_by_lead_time(long_forecast_multi_day, "MEAN").data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_collapse_by_lead_time_cube_list(
    long_forecast_multi_day, long_forecast_many_cubes
):
    """Check CubeList is made into an aggregatable cube and collapses by lead time."""
    calculated_cube = collapse.collapse(
        long_forecast_multi_day, "forecast_reference_time", "MEAN"
    )
    assert np.allclose(
        calculated_cube.data,
        collapse.collapse_by_lead_time(long_forecast_many_cubes, "MEAN").data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_collapse_by_lead_time_single_cube_percentile(long_forecast_multi_day):
    """Check Cube collapse by lead time with percentiles."""
    calculated_cube = collapse.collapse(
        long_forecast_multi_day,
        "forecast_reference_time",
        "PERCENTILE",
        additional_percent=75,
    )
    assert np.allclose(
        calculated_cube.data,
        collapse.collapse_by_lead_time(
            long_forecast_multi_day, "PERCENTILE", additional_percent=75
        ).data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_collapse_by_lead_time_single_cube_percentile_fail(long_forecast_multi_day):
    """Test fail by not setting additional percent."""
    with pytest.raises(ValueError):
        collapse.collapse_by_lead_time(long_forecast_multi_day, "PERCENTILE")


def test_collapse_by_lead_time_cube_list_percentile(
    long_forecast_multi_day, long_forecast_many_cubes
):
    """Check CubeList is made into an aggregatable cube and collapses by lead time with percentiles."""
    calculated_cube = collapse.collapse(
        long_forecast_multi_day,
        "forecast_reference_time",
        "PERCENTILE",
        additional_percent=75,
    )
    assert np.allclose(
        calculated_cube.data,
        collapse.collapse_by_lead_time(
            long_forecast_many_cubes, "PERCENTILE", additional_percent=75
        ).data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_collapse_by_validity_time(long_forecast_multi_day):
    """Reduce a dimension of a cube by validity time."""
    collapsed_cube = collapse.collapse_by_validity_time(long_forecast_multi_day, "MEAN")
    expected_cube = "<iris 'Cube' of air_temperature / (K) (time: 145; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_validity_time_cubelist(long_forecast_many_cubes):
    """Convert to cube and reduce a dimension by validity time."""
    collapsed_cube = collapse.collapse_by_validity_time(
        long_forecast_many_cubes, "MEAN"
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (time: 145; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_validity_time_percentile(long_forecast_multi_day):
    """Reduce by validity time with percentiles."""
    # Test successful collapsing by validity time.
    collapsed_cube = collapse.collapse_by_validity_time(
        long_forecast_multi_day, "PERCENTILE", additional_percent=[25, 75]
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (percentile_over_equalised_validity_time: 2; time: 145; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube


def test_collapse_by_validity_time_percentile_fail(long_forecast_multi_day):
    """Test not specifying additional percent fails."""
    with pytest.raises(ValueError):
        collapse.collapse_by_validity_time(long_forecast_multi_day, "PERCENTILE")


def test_collapse_by_validity_time_cubelist_percentile(long_forecast_many_cubes):
    """Convert to cube and reduce by validity time with percentiles."""
    collapsed_cube = collapse.collapse_by_validity_time(
        long_forecast_many_cubes, "PERCENTILE", additional_percent=[25, 75]
    )
    expected_cube = "<iris 'Cube' of air_temperature / (K) (percentile_over_equalised_validity_time: 2; time: 145; grid_latitude: 3; grid_longitude: 3)>"
    assert repr(collapsed_cube) == expected_cube



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Age of air operator.

The age of air diagnostic provides a qualtitative view of how old air is within
the domain, by calculating a back trajectory at each grid point at each lead time
to determine when air entered through the lateral boundary. This is useful for
diagnosing how quickly air ventilates the domain, depending on its size and the
prevailing meteorology.

The diagnostic uses the u, v and w components of wind, along with geopotential height to
perform the back trajectory. Data is first regridded to 0.5 degrees.

Note: the code here does not consider sub-grid transport, and only uses the postprocessed
velocity fields and geopotential height. Its applicability is for large-scale flow O(1000 km),
and not small scale flow where mixing is likely to play a larger role.
"""

import datetime
import logging
import multiprocessing
import os
import tempfile
from functools import partial
from math import asin, cos, radians, sin, sqrt

import numpy as np
from iris.cube import Cube
from scipy.ndimage import gaussian_filter

from CSET.operators._utils import get_cube_yxcoordname


def _calc_dist(coord_1, coord_2):
    """Calculate distance between two coordinate tuples.

    Arguments
    ----------
    coord_1: tuple
        A tuple containing (latitude, longitude) coordinate floats
    coord_2: tuple
        A tuple containing (latitude, longitude) coordinate floats

    Returns
    -------
    distance: float
        Distance between the two coordinate points in meters

    Notes
    -----
    The function uses the Haversine approximation to calculate distance in metres.

    """
    # Approximate radius of earth in m
    # Source: https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html
    radius = 6378000

    # Extract coordinates and convert to radians
    lat1 = radians(coord_1[0])
    lon1 = radians(coord_1[1])
    lat2 = radians(coord_2[0])
    lon2 = radians(coord_2[1])

    # Find out delta latitude, longitude
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    # Compute distance
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    distance = radius * c

    return distance


def _aoa_core(
    x_arr: np.ndarray,
    y_arr: np.ndarray,
    z_arr: np.ndarray,
    g_arr: np.ndarray,
    lats: np.ndarray,
    lons: np.ndarray,
    dt: int,
    plev_idx: int,
    timeunit: str,
    cyclic: bool,
    tmpdir: str,
    lon_pnt: int,
):
    """AOA multiprocessing core.

    Runs the core age of air code on a specific longitude point (all latitudes) for
    parallelisation.

    Arguments
    ---------
    x_arr: np.ndarray
        A numpy array containing x wind data.
    y_arr: np.ndarray
        A numpy array containing y wind data.
    z_arr: np.ndarray
        A numpy array containing w wind data.
    g_arr: np.ndarray
        A numpy array containing geopotential height data.
    lats: np.ndarray
        A numpy array containing latitude points.
    lons: np.ndarray
        A numpy array containing longitude points.
    dt: int
        Gap between time intervals
    plev_idx: int
        Index of pressure level requested to run back trajectories on.
    timeunit: str
        Units of time, currently only accepts 'hour'
    cyclic: bool
        Whether to wrap at east/west boundaries. See compute_ageofair for a fuller description.
    tmpdir: str
        Path to store intermediate data
    lon_pnt: int
        Longitude point to extract and run back trajectories on for parallelisation.
    """
    # Initialise empty array to store age of air for this latitude strip.
    ageofair_local = np.zeros((x_arr.shape[0], x_arr.shape[2]))
    logging.debug("Working on %s", lon_pnt)

    # Ignore leadtime 0 as this is trivial.
    for leadtime in range(1, x_arr.shape[0]):
        # Initialise leadtime slice with current leadtime.
        ageofair_local[leadtime, :] = leadtime * dt
        for lat_pnt in range(0, x_arr.shape[2]):
            # Gridpoint initialised as within LAM by construction.
            outside_lam = False

            # If final column, look at dist from prev column, otherwise look at next column.
            if lon_pnt == len(lons) - 1:
                ew_spacing = _calc_dist(
                    (lats[lat_pnt], lons[lon_pnt]), (lats[lat_pnt], lons[lon_pnt - 1])
                )
            else:
                ew_spacing = _calc_dist(
                    (lats[lat_pnt], lons[lon_pnt]), (lats[lat_pnt], lons[lon_pnt + 1])
                )

            # If final row, look at dist from row column, otherwise look at next row.
            if lat_pnt == len(lats) - 1:
                ns_spacing = _calc_dist(
                    (lats[lat_pnt], lons[lon_pnt]), (lats[lat_pnt - 1], lons[lon_pnt])
                )
            else:
                ns_spacing = _calc_dist(
                    (lats[lat_pnt], lons[lon_pnt]), (lats[lat_pnt + 1], lons[lon_pnt])
                )

            # Go through past timeslices
            for n in range(0, leadtime):
                # First step back, so we use i,j coords to find out parcel location
                # in terms of array point
                if n == 0:
                    x = lon_pnt
                    y = lat_pnt
                    z = plev_idx

                # Only seek preceding wind if its inside domain.
                if not outside_lam:
                    # Get vector profile at current time - nearest whole gridpoint.
                    u = x_arr[leadtime - n, int(z), int(y), int(x)]
                    v = y_arr[leadtime - n, int(z), int(y), int(x)]
                    w = z_arr[leadtime - n, int(z), int(y), int(x)]
                    g = g_arr[leadtime - n, int(z), int(y), int(x)]

                    # First, compute horizontal displacement using inverse of horizontal vector
                    # Convert m/s to m/[samplingrate]h, then m ->  model gridpoints
                    if timeunit == "hour":
                        du = ((u * 60 * 60 * dt) / ew_spacing) * -1.0
                        dv = ((v * 60 * 60 * dt) / ns_spacing) * -1.0
                        dz = (w * 60 * 60 * dt) * -1.0

                    # Get column of geopot height.
                    g_col = g_arr[(leadtime - n), :, int(y), int(x)]

                    # New geopotential height of parcel - store 'capacity' between timesteps as vertical motions smaller.
                    if n == 0:
                        new_g = g + dz
                        pre_g = new_g
                    else:
                        new_g = pre_g + dz

                    # Calculate which geopot level is closest to new geopot level.
                    z = np.argmin(np.abs(g_col - new_g))

                    # Update x,y location based on displacement. Z already updated
                    x = x + du
                    y = y + dv

                    # If it is now outside domain, then save age and don't process further with outside LAM flag.
                    # Support cyclic domains like K-SCALE, where x coord out of domain gets moved through dateline.
                    if cyclic:
                        if (
                            x < 0
                        ):  # as for example -0.3 would still be in domain, but x_arr.shape-0.3 would result in index error
                            x = x_arr.shape[3] + x  # wrap back around dateline
                        elif x >= x_arr.shape[3]:
                            x = x_arr.shape[3] - x
                    else:
                        if x < 0 or x >= x_arr.shape[3]:
                            ageofair_local[leadtime, lat_pnt] = n * dt
                            outside_lam = True

                    if y < 0 or y >= x_arr.shape[2]:
                        ageofair_local[leadtime, lat_pnt] = n * dt
                        outside_lam = True

    # Save 3d array containing age of air
    np.save(tmpdir + f"/aoa_frag_{lon_pnt:04d}.npy", ageofair_local)


def compute_ageofair(
    XWIND: Cube,
    YWIND: Cube,
    WWIND: Cube,
    GEOPOT: Cube,
    plev: int,
    cyclic: bool = False,
    multicore=True,
):
    """Compute back trajectories for a given forecast.

    This allows us to determine when air entered through the boundaries. This will run on all available
    lead-times, and thus return an age of air cube of shape ntime, nlat, nlon. It supports multiprocesing,
    by iterating over longitude, or if set as None, will run on a single core, which is easier for debugging.
    This function supports ensembles, where it will check if realization dimension exists and if so, loop
    over this axis.

    Arguments
    ----------
    XWIND: Cube
        An iris cube containing the x component of wind on pressure levels, on a 0p5 degree grid.
        Requires 4 dimensions, ordered time, pressure, latitude and longitude. Must contain at
        least 2 time points to compute back trajectory.
    YWIND: Cube
        An iris cube in the same format as XWIND.
    WWIND: Cube
        An iris cube in the same format as XWIND.
    GEOPOT: Cube
        An iris cube in the same format as XWIND.
    plev: int
        The pressure level of which to compute the back trajectory on. The function will search to
        see if this exists and if not, will raise an exception.
    cyclic: bool
        If cyclic is True, then the code will assume no east/west boundary and if a back trajectory
        reaches the boundary, it will emerge out of the other side. This option is useful for large
        domains such as the K-SCALE tropical channel, where there are only north/south boundaries in
        the domain.
    multicore: bool
        If true, split up age of air diagnostic to use multiple cores (defaults to number of cores available to the process), otherwise run
        using a single process, which is easier to debug if developing the code.

    Returns
    -------
    ageofair_cube: Cube
        An iris cube of the age of air data, with 3 dimensions (time, latitude, longitude).

    Notes
    -----
    The age of air diagnostic was used in Warner et al. (2023) [Warneretal2023]_ to identify the relative
    role of spin-up from initial conditions and lateral boundary conditions over tropical Africa to explore
    the impact of new data assimilation techniques. A further paper is currently in review ([Warneretal2024]_)
    which applies the diagnostic more widely to the Australian ACCESS convection-permitting models.

    References
    ----------
    .. [Warneretal2023] Warner, J.L., Petch, J., Short, C., Bain, C., 2023. Assessing the impact of an NWP warm-start
        system on model spin-up over tropical Africa. QJ, 149( 751), pp.621-636. doi:10.1002/qj.4429
    .. [Warneretal2024] Diagnosing lateral boundary spin-up in regional models using an age of air diagnostic
        James L. Warner, Charmaine N. Franklin, Belinda Roux, Shaun Cooper, Susan Rennie, Vinod
        Kumar.
        Submitted for Quarterly Journal of the Royal Meteorological Society.

    """
    # Set up temporary directory to store intermediate age of air slices.
    tmpdir = tempfile.TemporaryDirectory(dir=os.getenv("CYLC_TASK_WORK_DIR"))
    logging.info("Made tmpdir %s", tmpdir.name)

    # Check that all cubes are of same size (will catch different dimension orders too).
    if not XWIND.shape == YWIND.shape == WWIND.shape == GEOPOT.shape:
        raise ValueError("Cubes are not the same shape")

    # Get time units and assign for later
    if str(XWIND.coord("time").units).startswith("hours since "):
        timeunit = "hour"
    else:
        raise NotImplementedError("Unsupported time base")

    # Make data non-lazy to speed up code.
    logging.info("Making data non-lazy...")
    x_arr = XWIND.data
    y_arr = YWIND.data
    z_arr = WWIND.data
    g_arr = GEOPOT.data

    # Get coord points
    lat_name, lon_name = get_cube_yxcoordname(XWIND)
    lats = XWIND.coord(lat_name).points
    lons = XWIND.coord(lon_name).points
    time = XWIND.coord("time").points

    # Get time spacing of cube to determine whether the spacing in time is the
    # same throughout the cube. If not, then not supported.
    dt = XWIND.coord("time").points[1:] - XWIND.coord("time").points[:-1]
    if np.all(dt == dt[0]):
        dt = dt[0]
    else:
        raise NotImplementedError("Time intervals are not consistent")

    # Some logic to determine which index each axis is, and check for ensembles.
    dimension_mapping = {}
    for coord in XWIND.dim_coords:
        dim_index = XWIND.coord_dims(coord.name())[0]
        dimension_mapping[coord.name()] = dim_index

    if "realization" in dimension_mapping:
        ensemble_mode = True
        if dimension_mapping != {
            "realization": 0,
            "time": 1,
            "pressure": 2,
            lat_name: 3,
            lon_name: 4,
        }:
            raise ValueError(
                f"Dimension mapping not correct, ordered {dimension_mapping}"
            )
    else:
        ensemble_mode = False
        if dimension_mapping != {"time": 0, "pressure": 1, lat_name: 2, lon_name: 3}:
            raise ValueError(
                f"Dimension mapping not correct, ordered {dimension_mapping}"
            )

    # Smooth vertical velocity to 2sigma (standard for 0.5 degree).
    logging.info("Smoothing vertical velocity...")
    if ensemble_mode:
        z_arr = gaussian_filter(z_arr, 2, mode="nearest", axes=(3, 4))
    else:
        z_arr = gaussian_filter(z_arr, 2, mode="nearest", axes=(2, 3))

    # Get array index for user specified pressure level.
    if plev not in XWIND.coord("pressure").points:
        raise IndexError(f"Can't find plev {plev} in {XWIND.coord('pressure').points}")

    # Find corresponding pressure level index
    plev_idx = np.where(XWIND.coord("pressure").points == plev)[0][0]

    # Initialise cube containing age of air.
    if ensemble_mode:
        ageofair_cube = Cube(
            np.zeros(
                (
                    len(XWIND.coord("realization").points),
                    len(time),
                    len(lats),
                    len(lons),
                )
            ),
            long_name="age_of_air",
            units="hours",
            dim_coords_and_dims=[
                (XWIND.coord("realization"), 0),
                (XWIND.coord("time"), 1),
                (XWIND.coord(lat_name), 2),
                (XWIND.coord(lon_name), 3),
            ],
        )
    else:
        ageofair_cube = Cube(
            np.zeros((len(time), len(lats), len(lons))),
            long_name="age_of_air",
            units="hours",
            dim_coords_and_dims=[
                (XWIND.coord("time"), 0),
                (XWIND.coord(lat_name), 1),
                (XWIND.coord(lon_name), 2),
            ],
        )

    # Unix API for getting set of usable CPUs.
    # See https://docs.python.org/3/library/os.html#os.cpu_count
    if multicore:
        num_usable_cores = len(os.sched_getaffinity(0))
        # Use "spawn" method to avoid warnings before the default is changed in
        # python 3.14. See the (not very good) warning here:
        # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        mp_context = multiprocessing.get_context("spawn")
        pool = mp_context.Pool(num_usable_cores)

    logging.info("STARTING AOA DIAG...")
    start = datetime.datetime.now()

    # Main call for calculating age of air diagnostic
    if ensemble_mode:
        for e in range(0, len(XWIND.coord("realization").points)):
            logging.info(f"Working on member {e}")

            # Multiprocessing on each longitude slice
            func = partial(
                _aoa_core,
                np.copy(x_arr[e, :, :, :, :]),
                np.copy(y_arr[e, :, :, :, :]),
                np.copy(z_arr[e, :, :, :, :]),
                np.copy(g_arr[e, :, :, :, :]),
                lats,
                lons,
                dt,
                plev_idx,
                timeunit,
                cyclic,
                tmpdir.name,
            )
            if multicore:
                pool.map(func, range(0, XWIND.shape[4]))
            else:
                # Convert to list to ensure everything is processed.
                list(map(func, range(0, XWIND.shape[4])))

            for i in range(0, XWIND.shape[4]):
                file = f"{tmpdir.name}/aoa_frag_{i:04}.npy"
                ageofair_cube.data[e, :, :, i] = np.load(file)

    else:
        # Multiprocessing on each longitude slice
        func = partial(
            _aoa_core,
            np.copy(x_arr),
            np.copy(y_arr),
            np.copy(z_arr),
            np.copy(g_arr),
            lats,
            lons,
            dt,
            plev_idx,
            timeunit,
            cyclic,
            tmpdir.name,
        )
        if multicore:
            pool.map(func, range(0, XWIND.shape[3]))
        else:
            # Convert to list to ensure everything is processed.
            list(map(func, range(0, XWIND.shape[3])))

        for i in range(0, XWIND.shape[3]):
            file = f"{tmpdir.name}/aoa_frag_{i:04}.npy"
            ageofair_cube.data[:, :, i] = np.load(file)

    if multicore:
        # Wait for tasks to finish then clean up worker processes.
        pool.terminate()
        pool.join()

    # Verbose for time taken to run, and collate tmp ndarrays into final cube, and return
    logging.info(
        "AOA DIAG DONE, took %s s",
        (datetime.datetime.now() - start).total_seconds(),
    )

    # Clean tmpdir
    tmpdir.cleanup()

    return ageofair_cube



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fixtures defined in this module will be available to all tests.

https://docs.pytest.org/en/latest/reference/fixtures.html#conftest-py-sharing-fixtures-across-multiple-files
"""

import pytest

from CSET.operators import constraints, filters, read


@pytest.fixture()
def tmp_working_dir(tmp_path, monkeypatch):
    """Change the working directory for a test."""
    monkeypatch.chdir(tmp_path)
    return tmp_path


# Session scope fixtures, so the test data only has to be loaded from disk
# once, then make an in-memory copy whenever it is reused.
@pytest.fixture(scope="session")
def cubes_readonly():
    """Get an iris CubeList. NOT safe to modify."""
    return read.read_cubes("tests/test_data/air_temp.nc")


@pytest.fixture()
def cubes(cubes_readonly):
    """Get an iris CubeList. Safe to modify."""
    return cubes_readonly.copy()


@pytest.fixture(scope="session")
def cube_readonly(cubes_readonly):
    """Get an iris Cube. NOT safe to modify."""
    return filters.filter_cubes(
        cubes_readonly, constraints.generate_cell_methods_constraint([])
    )


@pytest.fixture()
def cube(cube_readonly):
    """Get an iris Cube. Safe to modify."""
    return cube_readonly.copy()


@pytest.fixture(scope="session")
def slammed_lfric_cube_readonly():
    """Get an iris Cube of slammed LFRic data. NOT safe to modify."""
    return read.read_cube("tests/test_data/slammed_lfric_air_temp.nc")


@pytest.fixture()
def slammed_lfric_cube(slammed_lfric_cube_readonly):
    """Get an iris Cube of slammed LFRic data. Safe to modify."""
    return slammed_lfric_cube_readonly.copy()


@pytest.fixture(scope="session")
def vertical_profile_cube_readonly():
    """Get a vertical profile Cube. It is NOT safe to modify."""
    return read.read_cube(
        "tests/test_data/air_temperature_vertical_profile_as_series.nc"
    )


@pytest.fixture()
def vertical_profile_cube(vertical_profile_cube_readonly):
    """Get a vertical profile Cube. It is safe to modify."""
    return vertical_profile_cube_readonly.copy()


@pytest.fixture(scope="session")
def histogram_cube_readonly():
    """Get a histogram Cube. It is NOT safe to modify."""
    from CSET.operators import read

    return read.read_cube(
        "tests/test_data/air_temperature_vertical_profile_as_series.nc"
    )


@pytest.fixture()
def histogram_cube(histogram_cube_readonly):
    """Get a histogram Cube."""
    return histogram_cube_readonly.copy()


@pytest.fixture(scope="session")
def regrid_rectilinear_cube_readonly():
    """Get a cube to test with. It is NOT safe to modify."""
    return read.read_cube(
        "tests/test_data/regrid/regrid_rectilinearGeogCS.nc", "surface_altitude"
    )


@pytest.fixture()
def regrid_rectilinear_cube(regrid_rectilinear_cube_readonly):
    """Get a cube to test with. It is safe to modify."""
    return regrid_rectilinear_cube_readonly.copy()


@pytest.fixture(scope="session")
def transect_source_cube_readonly():
    """Get a 3D cube to test with. It is NOT safe to modify."""
    return read.read_cube("tests/test_data/transect_test_umpl.nc")


@pytest.fixture()
def transect_source_cube(transect_source_cube_readonly):
    """Get a 3D cube to test with. It is safe to modify."""
    return transect_source_cube_readonly.copy()


@pytest.fixture()
def long_forecast():
    """Get long_forecast to run tests on."""
    return read.read_cube(
        "tests/test_data/long_forecast_air_temp_fcst_1.nc", "air_temperature"
    )


@pytest.fixture()
def long_forecast_multi_day():
    """Get long_forecast_multi_day to run tests on."""
    return read.read_cube(
        "tests/test_data/long_forecast_air_temp_multi_day.nc", "air_temperature"
    )


@pytest.fixture()
def long_forecast_many_cubes():
    """Get long_forecast_may_cubes to run tests on."""
    return read.read_cubes(
        "tests/test_data/long_forecast_air_temp_fcst_*.nc", "air_temperature"
    )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Subpackage contains all of CSET's operators."""

import inspect
import json
import logging
import os
from pathlib import Path

from iris import FUTURE

# Import operators here so they are exported for use by recipes.
import CSET.operators
from CSET._common import iter_maybe, parse_recipe
from CSET.operators import (
    ageofair,
    aggregate,
    collapse,
    constraints,
    convection,
    ensembles,
    filters,
    mesoscale,
    misc,
    plot,
    read,
    regrid,
    transect,
    write,
)

# Exported operators & functions to use elsewhere.
__all__ = [
    "ageofair",
    "aggregate",
    "collapse",
    "constraints",
    "convection",
    "ensembles",
    "execute_recipe",
    "filters",
    "get_operator",
    "mesoscale",
    "misc",
    "plot",
    "read",
    "regrid",
    "transect",
    "write",
]

# Stop iris giving a warning whenever it loads something.
FUTURE.datum_support = True
# Stop iris giving a warning whenever it saves something.
FUTURE.save_split_attrs = True
# Accept microsecond precision in iris times.
FUTURE.date_microseconds = True


def get_operator(name: str):
    """Get an operator by its name.

    Parameters
    ----------
    name: str
        The name of the desired operator.

    Returns
    -------
    function
        The named operator.

    Raises
    ------
    ValueError
        If name is not an operator.

    Examples
    --------
    >>> CSET.operators.get_operator("read.read_cubes")
    <function read_cubes at 0x7fcf9353c8b0>
    """
    logging.debug("get_operator(%s)", name)
    try:
        name_sections = name.split(".")
        operator = CSET.operators
        for section in name_sections:
            operator = getattr(operator, section)
        if callable(operator):
            return operator
        else:
            raise AttributeError
    except (AttributeError, TypeError) as err:
        raise ValueError(f"Unknown operator: {name}") from err


def _write_metadata(recipe: dict):
    """Write a meta.json file in the CWD."""
    # TODO: Investigate whether we might be better served by an SQLite database.
    metadata = recipe.copy()
    # Remove steps, as not needed, and might contain non-serialisable types.
    metadata.pop("steps", None)
    with open("meta.json", "wt", encoding="UTF-8") as fp:
        json.dump(metadata, fp, indent=2)
    os.sync()
    # Stat directory to force NFS to synchronise metadata.
    os.stat(Path.cwd())


def _step_parser(step: dict, step_input: any) -> str:
    """Execute a recipe step, recursively executing any sub-steps."""
    logging.debug("Executing step: %s", step)
    kwargs = {}
    for key in step.keys():
        if key == "operator":
            operator = get_operator(step["operator"])
            logging.info("operator: %s", step["operator"])
        elif isinstance(step[key], dict) and "operator" in step[key]:
            logging.debug("Recursing into argument: %s", key)
            kwargs[key] = _step_parser(step[key], step_input)
        else:
            kwargs[key] = step[key]
    logging.debug("args: %s", kwargs)
    logging.debug("step_input: %s", step_input)
    # If first argument of operator is explicitly defined, use that rather
    # than step_input. This is known through introspection of the operator.
    first_arg = next(iter(inspect.signature(operator).parameters.keys()))
    logging.debug("first_arg: %s", first_arg)
    if first_arg not in kwargs:
        logging.debug("first_arg not in kwargs, using step_input.")
        return operator(step_input, **kwargs)
    else:
        logging.debug("first_arg in kwargs.")
        return operator(**kwargs)


def _run_steps(
    recipe,
    steps,
    step_input,
    output_directory: Path,
    style_file: Path = None,
    plot_resolution: int = None,
) -> None:
    """Execute the steps in a recipe."""
    original_working_directory = Path.cwd()
    # Make paths absolute.
    step_input = [
        path if path.startswith("/") else f"{original_working_directory}/{path}"
        for path in iter_maybe(step_input)
    ]
    try:
        os.chdir(output_directory)
        logger = logging.getLogger()
        diagnostic_log = logging.FileHandler(
            filename="CSET.log", mode="w", encoding="UTF-8"
        )
        diagnostic_log.addFilter(lambda record: record.levelno >= logging.INFO)
        diagnostic_log.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)s %(message)s")
        )
        logger.addHandler(diagnostic_log)
        # Create metadata file used by some steps.
        if style_file:
            recipe["style_file_path"] = str(style_file)
        if plot_resolution:
            recipe["plot_resolution"] = plot_resolution
        _write_metadata(recipe)
        # Execute the recipe.
        for step in steps:
            step_input = _step_parser(step, step_input)
        logging.info("Recipe output:\n%s", step_input)
    finally:
        os.chdir(original_working_directory)


def execute_recipe(
    recipe_yaml: Path | str,
    input_directories: list[str],
    output_directory: Path,
    recipe_variables: dict = None,
    style_file: Path = None,
    plot_resolution: int = None,
) -> None:
    """Parse and executes the steps from a recipe file.

    Parameters
    ----------
    recipe_yaml: Path | str
        Path to a file containing, or a string of, a recipe's YAML describing
        the operators that need running. If a Path is provided it is opened and
        read.
    input_directories: list[str]
        List of pathlike to directories containing input files. Each path is
        assumed to be to a different model.
    output_directory: Path
        Pathlike indicating desired location of output.
    recipe_variables: dict, optional
        Dictionary of variables for the recipe.
    style_file: Path, optional
        Path to a style file.
    plot_resolution: int, optional
        Resolution of plots in dpi.

    Raises
    ------
    FileNotFoundError
        The recipe or input file cannot be found.
    FileExistsError
        The output directory as actually a file.
    ValueError
        The recipe is not well formed.
    TypeError
        The provided recipe is not a stream or Path.
    """
    recipe = parse_recipe(recipe_yaml, recipe_variables)
    # Create output directory.
    try:
        output_directory.mkdir(parents=True, exist_ok=True)
    except (FileExistsError, NotADirectoryError) as err:
        logging.error("Output directory is a file. %s", output_directory)
        raise err
    steps = recipe["steps"]
    _run_steps(
        recipe, steps, input_directories, output_directory, style_file, plot_resolution
    )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for age of air operator."""

import iris
import iris.cube
import numpy as np
import pytest

from CSET.operators import ageofair, read


@pytest.fixture()
def xwind() -> iris.cube.Cube:
    """Get regridded xwind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_rgd.nc", "x_wind")


@pytest.fixture()
def ywind() -> iris.cube.Cube:
    """Get regridded ywind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_rgd.nc", "y_wind")


@pytest.fixture()
def wwind() -> iris.cube.Cube:
    """Get regridded wwind to run tests on."""
    return read.read_cube(
        "tests/test_data/ageofair/aoa_in_rgd.nc", "upward_air_velocity"
    )


@pytest.fixture()
def geopot() -> iris.cube.Cube:
    """Get regridded geopotential height to run tests on."""
    return read.read_cube(
        "tests/test_data/ageofair/aoa_in_rgd.nc", "geopotential_height"
    )


@pytest.fixture()
def ens_regridded() -> iris.cube.CubeList:
    """Get regridded ensemble cubelist to run tests on."""
    return read.read_cubes("tests/test_data/ageofair/aoa_in_ens.nc")


@pytest.fixture()
def ens_regridded_out() -> iris.cube.Cube:
    """Get age of air ensemble output to check results are consistent."""
    return read.read_cube("tests/test_data/ageofair/aoa_out_ens.nc")


def test_calc_dist():
    """Test distance calculated from calc_dist in age of air calculation.

    Allow a tolerance of 20km (as we expect some error).
    """
    # London and Johannesburg coordinates to 2 decimal places.
    london_coords = (51.51, -0.13)
    johanbg_coords = (-26.21, 28.03)

    dist = ageofair._calc_dist(london_coords, johanbg_coords)
    actual_distance = 9068670  # Air line according to Google?!

    assert np.allclose(dist, actual_distance, rtol=1e-06, atol=20000)


def test_aoa_nocyclic(xwind, ywind, wwind, geopot):
    """Test case when not cyclic."""
    assert np.allclose(
        ageofair.compute_ageofair(
            xwind, ywind, wwind, geopot, plev=500, cyclic=False
        ).data,
        read.read_cube("tests/test_data/ageofair/aoa_out_nocyclic.nc").data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_aoa_cyclic_parallel_processing(xwind, ywind, wwind, geopot):
    """Test case when cyclic with parallel processing."""
    assert np.allclose(
        ageofair.compute_ageofair(
            xwind,
            ywind,
            wwind,
            geopot,
            plev=500,
            cyclic=True,
            multicore=True,
        ).data,
        read.read_cube("tests/test_data/ageofair/aoa_out_cyclic.nc").data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_aoa_cyclic_single_process(xwind, ywind, wwind, geopot):
    """Test case when cyclic with no parallelisation."""
    assert np.allclose(
        ageofair.compute_ageofair(
            xwind,
            ywind,
            wwind,
            geopot,
            plev=500,
            cyclic=True,
            multicore=False,
        ).data,
        read.read_cube("tests/test_data/ageofair/aoa_out_cyclic.nc").data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_aoa_mismatched_size(xwind, ywind, wwind, geopot):
    """Mismatched array size raises error."""
    ywind = ywind[:, :, 1:, :]
    with pytest.raises(ValueError):
        ageofair.compute_ageofair(xwind, ywind, wwind, geopot, plev=500, cyclic=True)


def test_aoa_timefreq(xwind, ywind, wwind, geopot):
    """Variable time intervals raises NotImplemented error."""
    with pytest.raises(NotImplementedError):
        ageofair.compute_ageofair(
            xwind[[1, 2, 4, 5], :, :, :],
            ywind[[1, 2, 4, 5], :, :, :],
            wwind[[1, 2, 4, 5], :, :, :],
            geopot[[1, 2, 4, 5], :, :, :],
            plev=500,
            cyclic=True,
        )


def test_aoa_timeunits(xwind, ywind, wwind, geopot):
    """Time intervals that are not hourly raises NotImplemented error."""
    xwind.coord("time").units = "days since 1970-01-01 00:00:00"
    with pytest.raises(NotImplementedError):
        ageofair.compute_ageofair(xwind, ywind, wwind, geopot, plev=500, cyclic=True)


def test_aoa_plevreq(xwind, ywind, wwind, geopot):
    """Pressure level requested that doesn't exist raises Index error."""
    with pytest.raises(IndexError):
        ageofair.compute_ageofair(xwind, ywind, wwind, geopot, plev=123, cyclic=True)


def test_aoa_ens(ens_regridded, ens_regridded_out):
    """Test case including vertical velocity and not cyclic for ensemble data."""
    assert np.allclose(
        ageofair.compute_ageofair(
            ens_regridded.extract("x_wind")[0],
            ens_regridded.extract("y_wind")[0],
            ens_regridded.extract("upward_air_velocity")[0],
            ens_regridded.extract("geopotential_height")[0],
            plev=200,
            cyclic=False,
            multicore=False,
        ).data,
        ens_regridded_out.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_aoa_ens_multicore(ens_regridded, ens_regridded_out):
    """Test case with ensembles ensuring that single core and multicore produce identical values."""
    assert np.allclose(
        ageofair.compute_ageofair(
            ens_regridded.extract("x_wind")[0],
            ens_regridded.extract("y_wind")[0],
            ens_regridded.extract("upward_air_velocity")[0],
            ens_regridded.extract("geopotential_height")[0],
            plev=200,
            cyclic=False,
            multicore=True,
        ).data,
        ens_regridded_out.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_aoa_misordered_dims(xwind, ywind, wwind, geopot):
    """Dimensions in input cubes do not match expected ordering."""
    xwind.transpose([0, 1, 3, 2])
    ywind.transpose([0, 1, 3, 2])
    wwind.transpose([0, 1, 3, 2])
    geopot.transpose([0, 1, 3, 2])
    with pytest.raises(ValueError):
        ageofair.compute_ageofair(
            xwind,
            ywind,
            wwind,
            geopot,
            plev=500,
            cyclic=True,
        )


def test_aoa_misordered_dims_ens(ens_regridded):
    """Dimensions in input cubes do not match expected ordering."""
    xwind = ens_regridded.extract("x_wind")[0]
    ywind = ens_regridded.extract("y_wind")[0]
    wwind = ens_regridded.extract("upward_air_velocity")[0]
    geopot = ens_regridded.extract("geopotential_height")[0]
    xwind.transpose([0, 1, 2, 4, 3])
    ywind.transpose([0, 1, 2, 4, 3])
    wwind.transpose([0, 1, 2, 4, 3])
    geopot.transpose([0, 1, 2, 4, 3])
    with pytest.raises(ValueError):
        ageofair.compute_ageofair(
            xwind,
            ywind,
            wwind,
            geopot,
            plev=500,
            cyclic=True,
        )



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for common operator functionality across CSET."""

import iris
import iris.coords
import iris.cube
import pytest

import CSET.operators._utils as operator_utils


def test_missing_coord_get_cube_yxcoordname_x(regrid_rectilinear_cube):
    """Missing X coordinate raises error."""
    regrid_rectilinear_cube.remove_coord("grid_longitude")
    with pytest.raises(ValueError):
        operator_utils.get_cube_yxcoordname(regrid_rectilinear_cube)


def test_missing_coord_get_cube_yxcoordname_y(regrid_rectilinear_cube):
    """Missing Y coordinate raises error."""
    regrid_rectilinear_cube.remove_coord("grid_longitude")
    with pytest.raises(ValueError):
        operator_utils.get_cube_yxcoordname(regrid_rectilinear_cube)


def test_get_cube_yxcoordname(regrid_rectilinear_cube):
    """Check that function returns tuple containing horizontal dimension names."""
    assert (operator_utils.get_cube_yxcoordname(regrid_rectilinear_cube)) == (
        "grid_latitude",
        "grid_longitude",
    )


def test_is_transect_multiple_spatial_coords(regrid_rectilinear_cube):
    """Check that function returns False as more than one spatial map coord."""
    assert not operator_utils.is_transect(regrid_rectilinear_cube)


def test_is_transect_no_vertical_coord(transect_source_cube):
    """Check that function returns False as no vertical coord found."""
    # Retain only time and latitude coordinate, so it passes the first spatial coord test.
    transect_source_cube_slice = transect_source_cube[:, 0, :, 0]
    assert not operator_utils.is_transect(transect_source_cube_slice)


def test_is_transect_correct_coord(transect_source_cube):
    """Check that function returns True as one map and vertical coord found."""
    # Retain only time and latitude coordinate, so it passes the first spatial coord test.
    transect_source_cube_slice = transect_source_cube[:, :, :, 0]
    assert operator_utils.is_transect(transect_source_cube_slice)


def test_is_spatialdim_false(transect_source_cube):
    """Check that is spatial test returns false if cube does not contain spatial coordinates."""
    cube = transect_source_cube[:, :, 0, 0]  # Remove spatial dimcoords.
    assert not operator_utils.is_spatialdim(cube)


def test_is_spatialdim_true(transect_source_cube):
    """Check that is spatial test returns true if cube contains spatial coordinates."""
    assert operator_utils.is_spatialdim(transect_source_cube)


def test_fully_equalise_attributes_remove_unique_attributes():
    """Check unique attributes are removed."""
    original = iris.cube.Cube(
        [], var_name="variable", attributes={"shared_attribute": 1}
    )
    c1 = original.copy()
    c2 = original.copy()
    c2.attributes["unique_attribute"] = 1

    fixed_cubes = operator_utils.fully_equalise_attributes([c1, c2])
    for cube in fixed_cubes:
        assert cube == original


def test_fully_equalise_attributes_remove_differing_attributes():
    """Check attributes with different values are removed."""
    original = iris.cube.Cube(
        [], var_name="variable", attributes={"shared_attribute": 1}
    )
    c1 = original.copy()
    c2 = original.copy()
    c2.attributes["shared_attribute"] = 2

    fixed_cubes = operator_utils.fully_equalise_attributes([c1, c2])
    for cube in fixed_cubes:
        assert "shared_attribute" not in cube.attributes


def test_fully_equalise_attributes_remove_unique_coords():
    """Check unique coordinates are removed."""
    foo_coord = iris.coords.DimCoord([0], var_name="foo")
    bar_coord = iris.coords.AuxCoord([0], var_name="bar")

    original = iris.cube.Cube(
        [0], var_name="variable", dim_coords_and_dims=[(foo_coord, 0)]
    )
    c1 = original.copy()
    c2 = original.copy()
    c2.add_aux_coord(bar_coord)

    fixed_cubes = operator_utils.fully_equalise_attributes([c1, c2])
    for cube in fixed_cubes:
        assert cube.coord("foo")
        assert not cube.coords("bar")


def test_fully_equalise_attributes_equalise_coords():
    """Check differing coordinates are equalised."""
    foo_coord = iris.coords.DimCoord(
        [0], var_name="foo", attributes={"shared_attribute": 1}
    )
    original = iris.cube.Cube(
        [0], var_name="variable", dim_coords_and_dims=[(foo_coord, 0)]
    )
    c1 = original.copy()
    c2 = original.copy()
    c2.coord("foo").attributes["shared_attribute"] = 2

    fixed_cubes = operator_utils.fully_equalise_attributes([c1, c2])
    for cube in fixed_cubes:
        assert "shared_attribute" not in cube.coord("foo").attributes


def test_is_time_aggregatable_False(long_forecast):
    """Check that a cube that is not time aggregatable returns False."""
    assert not operator_utils.is_time_aggregatable(long_forecast)


def test_is_time_aggregatable(long_forecast_multi_day):
    """Check that a time aggregatable cube returns True."""
    assert operator_utils.is_time_aggregatable(long_forecast_multi_day)



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualise recipe into a graph."""

import logging
import subprocess
import sys
import tempfile
from pathlib import Path
from uuid import uuid4

import pygraphviz

from CSET._common import parse_recipe


def save_graph(
    recipe_file: Path | str,
    save_path: Path = None,
    auto_open: bool = False,
    detailed: bool = False,
):
    """
    Draws out the graph of a recipe, and saves it to a file.

    Parameters
    ----------
    recipe_file: Path | str
        The recipe to be graphed.

    save_path: Path
        Path where to save the generated image. Defaults to a temporary file.

    auto_open: bool
        Whether to automatically open the graph with the default image viewer.

    detailed: bool
        Whether to include operator arguments on the graph.

    Raises
    ------
    ValueError
        Recipe is invalid.
    """
    recipe = parse_recipe(recipe_file)
    if save_path is None:
        save_path = Path(f"{tempfile.gettempdir()}/{uuid4()}.svg")

    def step_parser(step: dict, prev_node: str) -> str:
        """Parse recipe to add nodes to graph and link them with edges."""
        logging.debug("Executing step: %s", step)
        node = str(uuid4())
        graph.add_node(node, label=step["operator"])
        kwargs = {}
        for key in step.keys():
            if isinstance(step[key], dict) and "operator" in step[key]:
                logging.debug("Recursing into argument: %s", key)
                sub_node = step_parser(step[key], prev_node)
                graph.add_edge(sub_node, node)
            elif key != "operator":
                kwargs[key] = step[key]
        graph.add_edge(prev_node, node)

        if detailed:
            graph.get_node(node).attr["label"] = f"{step['operator']}\n" + "".join(
                f"<{key}: {kwargs[key]}>\n" for key in kwargs
            )
        return node

    graph = pygraphviz.AGraph(directed=True)

    prev_node = "START"
    graph.add_node(prev_node)
    try:
        for step in recipe["steps"]:
            prev_node = step_parser(step, prev_node)
    except KeyError as err:
        raise ValueError("Invalid recipe") from err

    graph.draw(save_path, format="svg", prog="dot")
    print(f"Graph rendered to {save_path}")

    if auto_open:
        try:
            # Stderr is redirected here to suppress gvfs-open deprecation warning.
            # See https://bugs.python.org/issue30219 for an example.
            subprocess.run(
                ("xdg-open", str(save_path)), check=True, stderr=subprocess.DEVNULL
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            # Using print rather than logging as this is run interactively.
            print(
                "Cannot automatically display graph. Specify an output with -o instead.",
                file=sys.stderr,
            )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for running CSET operator recipes."""

import json
from pathlib import Path

import pytest

import CSET.operators


def test_get_operator():
    """Happy case."""
    read_operator = CSET.operators.get_operator("read.read_cubes")
    assert callable(read_operator)


def test_get_operator_exception_missing():
    """Test exception for non-existent operators."""
    with pytest.raises(ValueError):
        CSET.operators.get_operator("non-existent.operator")


def test_get_operator_exception_type():
    """Test exception if wrong type provided."""
    with pytest.raises(ValueError):
        CSET.operators.get_operator(["Not", b"a", "string", 1])


def test_get_operator_exception_not_callable():
    """Test exception if operator isn't a function."""
    with pytest.raises(ValueError):
        CSET.operators.get_operator("misc.__doc__")


def test_execute_recipe(tmp_path: Path):
    """Execute recipe to test happy case (this is really an integration test)."""
    input_file = "tests/test_data/air_temp.nc"
    recipe = Path("tests/test_data/plot_instant_air_temp.yaml")
    CSET.operators.execute_recipe(recipe, input_file, tmp_path)


def test_execute_recipe_edge_cases(tmp_path: Path):
    """Test weird edge cases. Also tests data paths not being pathlib Paths."""
    input_file = "tests/test_data/air_temp.nc"
    recipe = Path("tests/test_data/noop_recipe.yaml")
    CSET.operators.execute_recipe(recipe, input_file, tmp_path)


def test_execute_recipe_invalid_output_dir(tmp_path: Path):
    """Exception raised if output directory can't be created."""
    recipe = '{"steps":[{"operator": misc.noop}]}'
    input_file = Path("tests/test_data/air_temp.nc")
    output_dir = tmp_path / "actually_a_file"
    output_dir.touch()
    with pytest.raises((FileExistsError, NotADirectoryError)):
        CSET.operators.execute_recipe(recipe, input_file, output_dir)


def test_run_steps_style_file_metadata_written(tmp_path: Path):
    """Style file path metadata written out."""
    style_file_path = "/test/style_file_path.json"
    CSET.operators._run_steps({}, [], "", tmp_path, Path(style_file_path))
    with open(tmp_path / "meta.json", "rb") as fp:
        metadata = json.load(fp)
    assert metadata["style_file_path"] == style_file_path


def test_run_steps_plot_resolution_metadata_written(tmp_path: Path):
    """Style file path metadata written out."""
    CSET.operators._run_steps({}, [], "", tmp_path, plot_resolution=72)
    with open(tmp_path / "meta.json", "rb") as fp:
        metadata = json.load(fp)
    assert metadata["plot_resolution"] == 72



#! /usr/bin/env python3

"""Write finished status to website front page.

Constructs the plot index, and does the final update to the workflow status on
the front page of the web interface.
"""

import json
import logging
import os
from pathlib import Path

from CSET._common import combine_dicts

logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"), format="%(asctime)s %(levelname)s %(message)s"
)


def construct_index():
    """Construct the plot index.

    Index should has the form ``{"Category Name": {"recipe_id": "Plot Name"}}``
    where ``recipe_id`` is the name of the plot's directory.
    """
    index = {}
    plots_dir = Path(os.environ["CYLC_WORKFLOW_SHARE_DIR"] + "/web/plots")
    # Loop over all directories, and append to index.
    # Only visits the directories directly under the plots directory.
    for directory in (d for d in plots_dir.iterdir() if d.is_dir()):
        try:
            with open(directory / "meta.json", "rt", encoding="UTF-8") as fp:
                plot_metadata = json.load(fp)
            record = {
                plot_metadata["category"]: {
                    directory.name: f"{plot_metadata['title']} {os.getenv('CYLC_TASK_CYCLE_POINT', '')}".strip()
                }
            }
        except FileNotFoundError:
            # Skip directories without metadata, as are likely not plots.
            logging.debug("No meta.json in %s, skipping.", directory)
            continue
        except (json.JSONDecodeError, KeyError, TypeError) as err:
            logging.error("%s is invalid, skipping.\n%s", directory / "meta.json", err)
            continue
        index = combine_dicts(index, record)

    with open(plots_dir / "index.json", "wt", encoding="UTF-8") as fp:
        json.dump(index, fp, indent=2)


def update_workflow_status():
    """Update the workflow status on the front page of the web interface."""
    web_dir = Path(os.environ["CYLC_WORKFLOW_SHARE_DIR"] + "/web")
    with open(web_dir / "status.html", "wt") as fp:
        fp.write("<p>Finished</p>\n")


def run():
    """Do the final steps to finish the website."""
    construct_index()
    update_workflow_status()



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing different diagnostics for convection.

The diagnostics are calculated from output from the Unified Model, although
precalculated values in the required input form may also be used.

"""

import copy
import logging
import warnings

import numpy as np


def cape_ratio(SBCAPE, MUCAPE, MUCIN, MUCIN_thresh=-75.0):
    r"""Ratio of two fields, one filtered to allow physical values to be output.

    Parameters
    ----------
    SBCAPE: Cube
        Surface-based convective available potential energy as calculated by the
        model. If using the UM please use STASH ``m01s20i114``
    MUCAPE: Cube
        Most-unstable convective available potential energy as calculated by the
        model. If using the UM please use STASH ``m01s20i112``
    MUCIN: Cube
        Most-unstable convective inhibition associated with the most-unstable
        ascent as calculated by the model. If using the UM please use STASH
        ``m01s20i113``
    MUCIN_thresh: float, optional, default is -75. J/kg.
        Threshold to filter the MUCAPE by values are realistically realisable.

    Returns
    -------
    Cube

    Notes
    -----
    This diagnostic is based on Clark et al. (2012) [Clarketal2012]_. It is
    based around the idea that for elevated convection the convective
    instability is not based at the surface. This utilises two flavours of CAPE:
    the surface-based CAPE (SBCAPE) and the most-unstable CAPE (MUCAPE). The
    MUCAPE is filtered by the MUCIN associated with that parcel's ascent to
    ensure that any CAPE can at least theoretically be released. The default
    value is set at -75 J/kg but it can be changes depending on location and
    users requirements.

    .. math:: 1 - (\frac{SBCAPE}{MUCAPE})

    The ratio is defined in this way such that if SBCAPE=MUCAPE the ratio will
    equal 1. If the ratio was reversed when MUCAPE exists and SBCAPE is zero the
    ratio would be undefined.

    The diagnostic varies smoothly between zero and unity. A value of 0 implies
    an environment is suitable for surface-based convection. A value of 1
    implies an environment is suitable for elevated convection. Values between
    imply transition convection with values closer to one imply elevated
    convection is more likely and values closer to zero implying that
    surface-based convection is more likely.

    Further details about this diagnostic for elevated convection identification
    can be found in Flack et al. (2023) [FlackCAPE2023]_.

    Expected applicability ranges: Convective-scale models will be noisier than
    parametrized models as they are more responsive to the convection, and thus
    it may be more sensible to view as a larger spatial average rather than on
    the native resolution.

    Interpretation notes: UM stash for CAPE and CIN are calculated at the end of
    the timestep. Therefore this diagnostic is applicable after precipitation
    has occurred, not before as is the usual interpretation of CAPE related
    diagnostics.

    References
    ----------
    .. [Clarketal2012] Clark, A. J., Kain J. S., Marsh P. T., Correia J., Xue
       M., and Kong F., (2012) "Forecasting tornado pathlengths using a
       three-dimensional object identification algorithm applied to
       convection-allowing forecasts." Weather and Forecasting, vol. 27,
       1090–1113, doi: 10.1175/WAF-D-11-00147.1
    .. [FlackCAPE2023] Flack, D.L.A., Lehnert, M., Lean, H.W., and Willington,
       S. (2023) "Characteristics of Diagnostics for Identifying Elevated
       Convection over the British Isles in a Convection-Allowing Model."
       Weather and Forecasting, vol. 30, 1079-1094, doi: 10.1175/WAF-D-22-0219.1

    Examples
    --------
    >>> CAPE_ratios=convection.cape_ratio(
            SBCAPE,MUCAPE,MUCIN)
    >>> iplt.pcolormesh(CAPE_ratios[0,:,:],cmap=mpl.cm.RdBu)
    >>> plt.gca().coastlines('10m')
    >>> plt.colorbar()
    >>> plt.clim(0,1)
    >>> plt.show()

    >>> CAPE_ratios=convection.cape_ratio(
            SBCAPE,MUCAPE,MUCIN,MUCIN_thresh=-1.5)
    >>> iplt.pcolormesh(CAPE_ratios[0,:,:],cmap=mpl.cm.RdBu)
    >>> plt.gca().coastlines('10m')
    >>> plt.clim(0,1)
    >>> plt.colorbar()
    >>> plt.show()
    """
    # Load in the data into the new arrays.
    MUCAPE_data = copy.deepcopy(MUCAPE.data)
    if isinstance(MUCAPE_data, np.ma.MaskedArray):
        MUCAPE_data = MUCAPE_data.filled(np.nan)
    # Remove all MUCAPE below MUCIN threshold.
    MUCAPE_data[MUCIN.data <= MUCIN_thresh] = np.nan
    with warnings.catch_warnings():
        # Ignore possible divide by zero warnings, as they are replaced by NaNs.
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        # Now calculate the main diagnostic.
        EC_Flagb = 1 - (SBCAPE.data / MUCAPE_data)
        if isinstance(EC_Flagb, np.ma.MaskedArray):
            EC_Flagb = EC_Flagb.filled(np.nan)
    # Filter to reduce NaN values and -inf values for plotting ease.
    # There are multiple types of NaN values so need to convert them all to same type.
    EC_Flagb[np.isnan(EC_Flagb)] = np.nan
    EC_Flagb[np.isinf(EC_Flagb)] = np.nan
    # Take the coordinates from an existing cube and replace the data.
    cape_ratio_cube = SBCAPE.copy()
    cape_ratio_cube.data = EC_Flagb
    # Rename and remove STASH code.
    cape_ratio_cube.var_name = "cape_ratio"
    cape_ratio_cube.attributes.pop("STASH", None)
    return cape_ratio_cube


def inflow_layer_properties(EIB, BLheight, Orography):
    r"""Filter to create a binary mask identifying elevated convection.

    Parameters
    ----------
    EIB: Cube
        Effective inflow layer base (precalculated or as identified by the
        model). If using the UM please use STASH ``m01s20i119``.
    BLheight: Cube
        Boundary layer height (precalculated or as identified by the model). If
        using the UM please use STASH ``m01s00i025``.
    Orography: Cube
        Model or actual orography, expected to be 2 dimensional. If 3 or 4
        dimensional cube given converts to 2 dimensions assuming static
        orography field in ensemble realization and time. If using the UM please
        use STASH ``m01s00i033``.

    Returns
    -------
    Cube

    Notes
    -----
    This diagnostic is based on the concept of an effective inflow layer. This
    concept was first introduced by Thompson et al. (2007) [Thompsonetal2007]_.
    The inflow layer defined the region of air that is most likely to be
    ingested into the convective event. It is defined by thresholding the CAPE
    and CIN values: CAPE > 100 J/kg and \|CIN\| < 250 J/kg.

    To turn this into a diagnostic for elevated convection the inflow layer base
    is filtered against the boundary layer height. The model orography is added
    to the boundary layer height to ensure reference height consistency as the
    BL height is defined above ground level and the inflow layer base is defined
    above sea level in the model output.

    .. math:: EIB > BLheight + Orography

    This is a binary diagnostic. It has a value of 0 to imply the environment is
    suitable for surface-based convection. It has a value of 1 to indicate the
    environment is suitable to produce elevated convection.

    Further details about this diagnostic for elevated convection identification
    can be found in Flack et al. (2023) [Flackinf2023]_.

    Expected applicability ranges: Convective-scale models will be noisier than
    parametrized models as they are more responsive to the convection, and thus
    it may be more sensible to view as a larger spatial average rather than at
    native resolution.

    Interpretation notes: The effective inflow layer base diagnostic from UM
    STASH is dependent upon the UM CAPE and CIN diagnostics. These diagnostics
    are calculated at the end of the timestep. Therefore this diagnostic is
    applicable after precipitation has occurred, not before as is the usual
    interpretation of CAPE related diagnostics.

    You might encounter warnings with the following text ``Orography assumed not
    to vary with ensemble member.`` or ``Orography assumed not to vary with time
    and ensemble member.`` these warnings are expected when the orography files
    are not 2-dimensional, and do not cause any problems unless ordering is not
    as expected.

    References
    ----------
    .. [Thompsonetal2007] Thompson, R. L. Mead, C. M., and Edwards, R., (2007)
       "Effective Storm-Relative Helicity and Bulk Shear in Supercell
       Thunderstorm Environments." Weather and Forecasting, vol. 22, 102-115,
       doi: 10.1175/WAF969.1
    .. [Flackinf2023] Flack, D.L.A., Lehnert, M., Lean, H.W., and Willington, S.
       (2023) "Characteristics of Diagnostics for Identifying Elevated
       Convection over the British Isles in a Convection-Allowing Model."
       Weather and Forecasting, vol. 30, 1079-1094, doi: 10.1175/WAF-D-22-0219.1

    Examples
    --------
    >>> Inflow_properties=convection.inflow_layer_properties(EIB,BLheight,Orography)
    >>> iplt.pcolormesh(Inflow_properties[0,:,:],cmap=mpl.cm.Purples)
    >>> plt.gca().coastlines('10m')
    >>> plt.colorbar()
    >>> plt.clim(0,1)
    >>> plt.show()

    """
    # Setup new array for output of the diagnostic.
    EC_Flagd = np.zeros(EIB.shape)
    # Check dimensions for Orography cube and replace with 2D array if not 2D.
    if Orography.ndim == 3:
        Orography = Orography.slices_over("realization").next()
        logging.warning("Orography assumed not to vary with ensemble member")
    elif Orography.ndim == 4:
        Orography = Orography.slices_over(("time", "realization")).next()
        logging.warning("Orography assumed not to vary with time or ensemble member. ")
    # Masked arrays are not respected, so convert masked values into NaNs.
    if isinstance(EIB.data, np.ma.MaskedArray):
        EIB.data = EIB.data.filled(np.nan)
    # Change points where Effective inflow layer base is larger than boundary
    # layer height to 1 implying elevated convection.
    EC_Flagd[EIB.data > (BLheight.data + Orography.data)] = 1.0
    # Take the coordinates from an existing cube and replace the data.
    inflow_properties_cube = EIB.copy()
    inflow_properties_cube.data = EC_Flagd
    # Rename and remove STASH code.
    inflow_properties_cube.var_name = "inflow_layer_properties"
    inflow_properties_cube.attributes.pop("STASH", None)
    return inflow_properties_cube



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to produce various kinds of plots."""

import fcntl
import functools
import importlib.resources
import json
import logging
import math
import sys
from typing import Literal

import iris
import iris.coords
import iris.cube
import iris.exceptions
import iris.plot as iplt
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from markdown_it import MarkdownIt

from CSET._common import (
    combine_dicts,
    get_recipe_metadata,
    iter_maybe,
    render_file,
    slugify,
)
from CSET.operators._utils import get_cube_yxcoordname, is_transect

# Use a non-interactive plotting backend.
mpl.use("agg")

############################
# Private helper functions #
############################


def _append_to_plot_index(plot_index: list) -> list:
    """Add plots into the plot index, returning the complete plot index."""
    with open("meta.json", "r+t", encoding="UTF-8") as fp:
        fcntl.flock(fp, fcntl.LOCK_EX)
        fp.seek(0)
        meta = json.load(fp)
        complete_plot_index = meta.get("plots", [])
        complete_plot_index = complete_plot_index + plot_index
        meta["plots"] = complete_plot_index
        fp.seek(0)
        fp.truncate()
        json.dump(meta, fp, indent=2)
    return complete_plot_index


def _check_single_cube(cube: iris.cube.Cube | iris.cube.CubeList) -> iris.cube.Cube:
    """Ensure a single cube is given.

    If a CubeList of length one is given that the contained cube is returned,
    otherwise an error is raised.

    Parameters
    ----------
    cube: Cube | CubeList
        The cube to check.

    Returns
    -------
    cube: Cube
        The checked cube.

    Raises
    ------
    TypeError
        If the input cube is not a Cube or CubeList of a single Cube.
    """
    if isinstance(cube, iris.cube.Cube):
        return cube
    if isinstance(cube, iris.cube.CubeList):
        if len(cube) == 1:
            return cube[0]
    raise TypeError("Must have a single cube", cube)


def _py312_importlib_resources_files_shim():
    """Importlib behaviour changed in 3.12 to avoid circular dependencies.

    This shim is needed until python 3.12 is our oldest supported version, after
    which it can just be replaced by directly using importlib.resources.files.
    """
    if sys.version_info.minor >= 12:
        files = importlib.resources.files()
    else:
        import CSET.operators

        files = importlib.resources.files(CSET.operators)
    return files


def _make_plot_html_page(plots: list):
    """Create a HTML page to display a plot image."""
    # Debug check that plots actually contains some strings.
    assert isinstance(plots[0], str)

    # Load HTML template file.
    operator_files = _py312_importlib_resources_files_shim()
    template_file = operator_files.joinpath("_plot_page_template.html")

    # Get some metadata.
    meta = get_recipe_metadata()
    title = meta.get("title", "Untitled")
    description = MarkdownIt().render(meta.get("description", "*No description.*"))

    # Prepare template variables.
    variables = {
        "title": title,
        "description": description,
        "initial_plot": plots[0],
        "plots": plots,
        "title_slug": slugify(title),
    }

    # Render template.
    html = render_file(template_file, **variables)

    # Save completed HTML.
    with open("index.html", "wt", encoding="UTF-8") as fp:
        fp.write(html)


@functools.cache
def _load_colorbar_map(user_colorbar_file: str = None) -> dict:
    """Load the colorbar definitions from a file.

    This is a separate function to make it cacheable.
    """
    colorbar_file = _py312_importlib_resources_files_shim().joinpath(
        "_colorbar_definition.json"
    )
    with open(colorbar_file, "rt", encoding="UTF-8") as fp:
        colorbar = json.load(fp)

    logging.debug("User colour bar file: %s", user_colorbar_file)
    override_colorbar = {}
    if user_colorbar_file:
        try:
            with open(user_colorbar_file, "rt", encoding="UTF-8") as fp:
                override_colorbar = json.load(fp)
        except FileNotFoundError:
            logging.warning("Colorbar file does not exist. Using default values.")

    # Overwrite values with the user supplied colorbar definition.
    colorbar = combine_dicts(colorbar, override_colorbar)
    return colorbar


def _colorbar_map_levels(cube: iris.cube.Cube):
    """Get an appropriate colorbar for the given cube.

    For the given variable the appropriate colorbar is looked up from a
    combination of the built-in CSET colorbar definitions, and any user supplied
    definitions. As well as varying on variables, these definitions may also
    exist for specific pressure levels to account for variables with
    significantly different ranges at different heights.

    Parameters
    ----------
    cube: Cube
        Cube of variable for which the colorbar information is desired.

    Returns
    -------
    cmap:
        Matplotlib colormap.
    levels:
        List of levels to use for plotting. For continuous plots the min and max
        should be taken as the range.
    norm:
        BoundaryNorm information.
    """
    # Grab the colorbar file from the recipe global metadata.
    user_colorbar_file = get_recipe_metadata().get("style_file_path", None)
    colorbar = _load_colorbar_map(user_colorbar_file)

    try:
        # We assume that pressure is a scalar coordinate here.
        pressure_level_raw = cube.coord("pressure").points[0]
        # Ensure pressure_level is a string, as it is used as a JSON key.
        pressure_level = str(int(pressure_level_raw))
    except iris.exceptions.CoordinateNotFoundError:
        pressure_level = None

    # First try long name, then standard name, then var name. This order is used
    # as long name is the one we correct between models, so it most likely to be
    # consistent.
    varnames = filter(None, [cube.long_name, cube.standard_name, cube.var_name])
    for varname in varnames:
        # Get the colormap for this variable.
        try:
            cmap = mpl.colormaps[colorbar[varname]["cmap"]]
            if pressure_level is None:
                var_colorbar = colorbar[varname]
            else:
                # If pressure level is specified for cube use a pressure-level
                # specific colorbar, if one exists.
                try:
                    var_colorbar = colorbar[varname]["pressure_levels"][pressure_level]
                except KeyError:
                    logging.warning(
                        "%s has no colorbar definition for pressure level %s.",
                        varname,
                        pressure_level,
                    )
                    # Fallback to variable default.
                    var_colorbar = colorbar[varname]

            # Get the colorbar levels for this variable.
            try:
                levels = var_colorbar["levels"]
                # Use discrete bins when levels are specified, rather than a
                # smooth range.
                norm = mpl.colors.BoundaryNorm(levels, ncolors=cmap.N)
                logging.debug("Using levels for %s colorbar.", varname)
            except KeyError:
                # Get the range for this variable.
                vmin, vmax = var_colorbar["min"], var_colorbar["max"]
                logging.debug("Using min and max for %s colorbar.", varname)
                # Calculate levels from range.
                levels = np.linspace(vmin, vmax, 51)
                norm = None
            return cmap, levels, norm
        except KeyError:
            logging.debug("Cube name %s has no colorbar definition.", varname)
            # Retry with next name.
            continue

    # Default if no varnames match.
    logging.warning("No colorbar definition exists for %s.", cube.name())
    cmap, levels, norm = mpl.colormaps["viridis"], None, None
    return cmap, levels, norm


def _get_plot_resolution() -> int:
    """Get resolution of rasterised plots in pixels per inch."""
    return get_recipe_metadata().get("plot_resolution", 100)


def _plot_and_save_spatial_plot(
    cube: iris.cube.Cube,
    filename: str,
    title: str,
    method: Literal["contourf", "pcolormesh"],
    **kwargs,
):
    """Plot and save a spatial plot.

    Parameters
    ----------
    cube: Cube
        2 dimensional (lat and lon) Cube of the data to plot.
    filename: str
        Filename of the plot to write.
    title: str
        Plot title.
    method: "contourf" | "pcolormesh"
        The plotting method to use.
    """
    # Setup plot details, size, resolution, etc.
    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")

    # Specify the color bar
    cmap, levels, norm = _colorbar_map_levels(cube)

    if method == "contourf":
        # Filled contour plot of the field.
        plot = iplt.contourf(cube, cmap=cmap, levels=levels, norm=norm)
    elif method == "pcolormesh":
        try:
            vmin = min(levels)
            vmax = max(levels)
        except TypeError:
            vmin, vmax = None, None
        # pcolormesh plot of the field.
        plot = iplt.pcolormesh(cube, cmap=cmap, norm=norm, vmin=vmin, vmax=vmax)
    else:
        raise ValueError(f"Unknown plotting method: {method}")

    # Using pyplot interface here as we need iris to generate a cartopy GeoAxes.
    axes = plt.gca()

    # Add coastlines if cube contains x and y map coordinates.
    # If is spatial map, fix extent to keep plot tight.
    try:
        lat_axis, lon_axis = get_cube_yxcoordname(cube)
        axes.coastlines(resolution="10m")
        axes.set_extent(
            [
                np.min(cube.coord(lon_axis).points),
                np.max(cube.coord(lon_axis).points),
                np.min(cube.coord(lat_axis).points),
                np.max(cube.coord(lat_axis).points),
            ]
        )
    except ValueError:
        # Skip if no x and y map coordinates.
        pass

    # Check to see if transect, and if so, adjust y axis.
    if is_transect(cube):
        if "pressure" in [coord.name() for coord in cube.coords()]:
            axes.invert_yaxis()
            axes.set_yscale("log")
            axes.set_ylim(1100, 100)
        # If both model_level_number and level_height exists, iplt can construct
        # plot as a function of height above orography (NOT sea level).
        elif {"model_level_number", "level_height"}.issubset(
            {coord.name() for coord in cube.coords()}
        ):
            axes.set_yscale("log")

        axes.set_title(
            f"{title}\n"
            f"Start Lat: {cube.attributes['transect_coords'].split('_')[0]}"
            f" Start Lon: {cube.attributes['transect_coords'].split('_')[1]}"
            f" End Lat: {cube.attributes['transect_coords'].split('_')[2]}"
            f" End Lon: {cube.attributes['transect_coords'].split('_')[3]}",
            fontsize=16,
        )

    else:
        # Add title.
        axes.set_title(title, fontsize=16)

    # Add watermark with min/max/mean. Currently not user togglable.
    # In the bbox dictionary, fc and ec are hex colour codes for grey shade.
    axes.annotate(
        f"Min: {np.min(cube.data):.3g} Max: {np.max(cube.data):.3g} Mean: {np.mean(cube.data):.3g}",
        xy=(1, -0.05),
        xycoords="axes fraction",
        xytext=(-5, 5),
        textcoords="offset points",
        ha="right",
        va="bottom",
        size=11,
        bbox=dict(boxstyle="round", fc="#cccccc", ec="#808080", alpha=0.9),
    )

    # Add colour bar.
    cbar = fig.colorbar(plot)
    cbar.set_label(label=f"{cube.name()} ({cube.units})", size=20)

    # Save plot.
    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved spatial plot to %s", filename)
    plt.close(fig)


def _plot_and_save_postage_stamp_spatial_plot(
    cube: iris.cube.Cube,
    filename: str,
    stamp_coordinate: str,
    title: str,
    method: Literal["contourf", "pcolormesh"],
    **kwargs,
):
    """Plot postage stamp spatial plots from an ensemble.

    Parameters
    ----------
    cube: Cube
        Iris cube of data to be plotted. It must have the stamp coordinate.
    filename: str
        Filename of the plot to write.
    stamp_coordinate: str
        Coordinate that becomes different plots.
    method: "contourf" | "pcolormesh"
        The plotting method to use.

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions.
    """
    # Use the smallest square grid that will fit the members.
    grid_size = int(math.ceil(math.sqrt(len(cube.coord(stamp_coordinate).points))))

    fig = plt.figure(figsize=(10, 10))

    # Specify the color bar
    cmap, levels, norm = _colorbar_map_levels(cube)

    # Make a subplot for each member.
    for member, subplot in zip(
        cube.slices_over(stamp_coordinate), range(1, grid_size**2 + 1), strict=False
    ):
        # Implicit interface is much easier here, due to needing to have the
        # cartopy GeoAxes generated.
        plt.subplot(grid_size, grid_size, subplot)
        if method == "contourf":
            # Filled contour plot of the field.
            plot = iplt.contourf(member, cmap=cmap, levels=levels, norm=norm)
        elif method == "pcolormesh":
            try:
                vmin = min(levels)
                vmax = max(levels)
            except TypeError:
                vmin, vmax = None, None
            # pcolormesh plot of the field.
            plot = iplt.pcolormesh(member, cmap=cmap, norm=norm, vmin=vmin, vmax=vmax)
        else:
            raise ValueError(f"Unknown plotting method: {method}")
        ax = plt.gca()
        ax.set_title(f"Member #{member.coord(stamp_coordinate).points[0]}")
        ax.set_axis_off()

        # Add coastlines if cube contains x and y map coordinates.
        # If is spatial map, fix extent to keep plot tight.
        try:
            lat_axis, lon_axis = get_cube_yxcoordname(cube)
            ax.coastlines(resolution="10m")
            ax.set_extent(
                [
                    np.min(cube.coord(lon_axis).points),
                    np.max(cube.coord(lon_axis).points),
                    np.min(cube.coord(lat_axis).points),
                    np.max(cube.coord(lat_axis).points),
                ]
            )
        except ValueError:
            # Skip if no x and y map coordinates.
            pass

    # Put the shared colorbar in its own axes.
    colorbar_axes = fig.add_axes([0.15, 0.07, 0.7, 0.03])
    colorbar = fig.colorbar(plot, colorbar_axes, orientation="horizontal")
    colorbar.set_label(f"{cube.name()} / {cube.units}")

    # Overall figure title.
    fig.suptitle(title)

    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved contour postage stamp plot to %s", filename)
    plt.close(fig)


def _plot_and_save_line_series(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    coord: iris.coords.Coord,
    filename: str,
    title: str,
    **kwargs,
):
    """Plot and save a 1D line series.

    Parameters
    ----------
    cubes: Cube or CubeList
        Cube or CubeList containing the cubes to plot on the y-axis.
    coord: Coord
        Coordinate to plot on the x-axis.
    filename: str
        Filename of the plot to write.
    title: str
        Plot title.
    """
    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")

    for cube_iter in iter_maybe(cubes):
        iplt.plot(coord, cube_iter, "o-")

    # Get the current axes
    ax = plt.gca()

    # Add some labels and tweak the style.
    # check if cubes[0] works for single cube if not CubeList
    ax.set(
        xlabel=f"{coord.name()} / {coord.units}",
        ylabel=f"{cubes[0].name()} / {cubes[0].units}",
        title=title,
    )
    ax.ticklabel_format(axis="y", useOffset=False)
    ax.tick_params(axis="x", labelrotation=15)
    ax.autoscale()

    # Save plot.
    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved line plot to %s", filename)
    plt.close(fig)


def _plot_and_save_vertical_line_series(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    coord: iris.coords.Coord,
    filename: str,
    series_coordinate: str,
    title: str,
    vmin: float,
    vmax: float,
    **kwargs,
):
    """Plot and save a 1D line series in vertical.

    Parameters
    ----------
    cubes: Cube or CubeList
        1 dimensional Cube or CubeList of the data to plot on x-axis.
    coord: Coord
        Coordinate to plot on y-axis.
    filename: str
        Filename of the plot to write.
    series_coordinate: str
        Coordinate to use as vertical axis.
    title: str
        Plot title.
    vmin: float
        Minimum value for the x-axis.
    vmax: float
        Maximum value for the x-axis.
    """
    # plot the vertical pressure axis using log scale
    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")
    for cube in iter_maybe(cubes):
        iplt.plot(cube, coord, "o-")

    # Get the current axis
    ax = plt.gca()

    # Special handling for pressure level data.
    if series_coordinate == "pressure":
        # Invert y-axis and set to log scale.
        ax.invert_yaxis()
        ax.set_yscale("log")

        # Define y-ticks and labels for pressure log axis.
        y_tick_labels = [
            "1000",
            "850",
            "700",
            "500",
            "300",
            "200",
            "100",
            "50",
            "30",
            "20",
            "10",
        ]
        y_ticks = [1000, 850, 700, 500, 300, 200, 100, 50, 30, 20, 10]

        # Set y-axis limits and ticks.
        ax.set_ylim(1100, 100)

    # Test if series_coordinate is model level data. The UM data uses
    # model_level_number and lfric uses full_levels as coordinate.
    elif series_coordinate in ("model_level_number", "full_levels", "half_levels"):
        # Define y-ticks and labels for vertical axis.
        y_ticks = iter_maybe(cubes)[0].coord(series_coordinate).points
        y_tick_labels = [str(int(i)) for i in y_ticks]
        ax.set_ylim(min(y_ticks), max(y_ticks))

    ax.set_yticks(y_ticks)
    ax.set_yticklabels(y_tick_labels)

    # set x-axis limits
    ax.set_xlim(vmin, vmax)

    # Add some labels and tweak the style.
    ax.set(
        ylabel=f"{coord.name()} / {coord.units}",
        xlabel=f"{iter_maybe(cubes)[0].name()} / {iter_maybe(cubes)[0].units}",
        title=title,
    )
    ax.ticklabel_format(axis="x")
    ax.tick_params(axis="y")

    # Save plot.
    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved line plot to %s", filename)
    plt.close(fig)


def _plot_and_save_scatter_plot(
    cube_x: iris.cube.Cube | iris.cube.CubeList,
    cube_y: iris.cube.Cube | iris.cube.CubeList,
    filename: str,
    title: str,
    one_to_one: bool,
    **kwargs,
):
    """Plot and save a 2D scatter plot.

    Parameters
    ----------
    cube_x: Cube | CubeList
        1 dimensional Cube or CubeList of the data to plot on x-axis.
    cube_y: Cube | CubeList
        1 dimensional Cube or CubeList of the data to plot on y-axis.
    filename: str
        Filename of the plot to write.
    title: str
        Plot title.
    one_to_one: bool
        Whether a 1:1 line is plotted.
    """
    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")
    # plot the cube_x and cube_y 1D fields as a scatter plot. If they are CubeLists this ensures
    # to pair each cube from cube_x with the corresponding cube from cube_y, allowing to iterate
    # over the pairs simultaneously.

    # Ensure cube_x and cube_y are iterable
    cube_x_iterable = iter_maybe(cube_x)
    cube_y_iterable = iter_maybe(cube_y)

    for cube_x_iter, cube_y_iter in zip(cube_x_iterable, cube_y_iterable, strict=True):
        iplt.scatter(cube_x_iter, cube_y_iter)
    if one_to_one is True:
        plt.plot(
            [
                np.nanmin([np.nanmin(cube_y.data), np.nanmin(cube_x.data)]),
                np.nanmax([np.nanmax(cube_y.data), np.nanmax(cube_x.data)]),
            ],
            [
                np.nanmin([np.nanmin(cube_y.data), np.nanmin(cube_x.data)]),
                np.nanmax([np.nanmax(cube_y.data), np.nanmax(cube_x.data)]),
            ],
            "k",
            linestyle="--",
        )
    ax = plt.gca()

    # Add some labels and tweak the style.
    ax.set(
        xlabel=f"{cube_x[0].name()} / {cube_x[0].units}",
        ylabel=f"{cube_y[0].name()} / {cube_y[0].units}",
        title=title,
    )
    ax.ticklabel_format(axis="y", useOffset=False)
    ax.tick_params(axis="x", labelrotation=15)
    ax.autoscale()

    # Save plot.
    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved scatter plot to %s", filename)
    plt.close(fig)


def _plot_and_save_histogram_series(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    filename: str,
    title: str,
    vmin: float,
    vmax: float,
    histtype: str = "step",
    **kwargs,
):
    """Plot and save a histogram series.

    Parameters
    ----------
    cubes: Cube or CubeList
        2 dimensional Cube or CubeList of the data to plot as histogram.
        Plotting options are fixed:
        density=True, histtype='step',stacked=True to ensure that
        a probability density is plotted using matplotlib.pyplot.hist
        to plot the probability density so that the area under
        the histogram integrates to 1.
        stacked is set to True so the sum of the histograms is
        normalized to 1.
        ax.autoscale is switched off and the ylim range
        is preset as (0,1) to make figures comparable.
    filename: str
        Filename of the plot to write.
    title: str
        Plot title.
    vmin: float
        minimum for colorbar
    vmax: float
        maximum for colorbar
    histtype: str
        The type of histogram to plot. Options are "step" for a line
        histogram or "barstacked", "stepfilled". "Step" is the default option,
        but can be changed in the rose-suite.conf configuration.
    """
    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")
    ax = plt.gca()

    for cube in iter_maybe(cubes):
        # Easier to check title (where var name originates)
        # than seeing if long names exist etc.
        # Exception case, where distribution better fits log scales/bins.
        if "surface_microphysical_rainfall_rate" in title:
            # Usually in seconds but mm/hr more intuitive.
            cube.convert_units("kg m-2 h-1")
            bins = 10.0 ** (
                np.arange(-10, 27, 1) / 10.0
            )  # Suggestion from RMED toolbox.
            bins = np.insert(bins, 0, 0)
            ax.set_xscale("log")
            ax.set_yscale("log")
            vmin = 0
            vmax = 400  # Manually set vmin/vmax to override json derived value.
        else:
            bins = np.linspace(vmin, vmax, 51)

        # Reshape cube data into a single array to allow for a single histogram.
        # Otherwise we plot xdim histograms stacked.
        cube_data_1d = (cube.data).flatten()

    x, y = np.histogram(cube_data_1d, bins=bins, density=True)
    ax.plot(y[:-1], x, color="black", linewidth=2, marker="o", markersize=6)

    # Add some labels and tweak the style.
    ax.set(
        title=title,
        xlabel=f"{iter_maybe(cubes)[0].name()} / {iter_maybe(cubes)[0].units}",
        ylabel="normalised probability density",
        xlim=(vmin, vmax),
    )

    # Overlay grid-lines onto histogram plot.
    ax.grid(linestyle="--", color="grey", linewidth=1)

    # Save plot.
    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved line plot to %s", filename)
    plt.close(fig)


def _plot_and_save_postage_stamp_histogram_series(
    cube: iris.cube.Cube,
    filename: str,
    title: str,
    stamp_coordinate: str,
    vmin: float,
    vmax: float,
    histtype: str,
    **kwargs,
):
    """Plot and save postage (ensemble members) stamps for a histogram series.

    Parameters
    ----------
    cube: Cube
        2 dimensional Cube of the data to plot as histogram.
        Plotting options are fixed:
        density=True, histtype='bar', stacked=True to ensure that
        a probability density is plotted using matplotlib.pyplot.hist
        to plot the probability density so that the area under
        the histogram integrates to 1.
        stacked is set to True so the sum of the histograms is
        normalized to 1.
        ax.autoscale is switched off and the ylim range
        is preset as (0,1) to make figures comparable.
    filename: str
        Filename of the plot to write.
    title: str
        Plot title.
    stamp_coordinate: str
        Coordinate that becomes different plots.
    vmin: float
        minimum for pdf x-axis
    vmax: float
        maximum for pdf x-axis
    histtype: str
        The type of histogram to plot. Options are "step" for a line
        histogram or "barstacked", "stepfilled". "Step" is the default option,
        but can be changed in the rose-suite.conf configuration.

    """
    # Use the smallest square grid that will fit the members.
    grid_size = int(math.ceil(math.sqrt(len(cube.coord(stamp_coordinate).points))))

    fig = plt.figure(figsize=(10, 10), facecolor="w", edgecolor="k")
    # Make a subplot for each member.
    for member, subplot in zip(
        cube.slices_over(stamp_coordinate), range(1, grid_size**2 + 1), strict=False
    ):
        # Implicit interface is much easier here, due to needing to have the
        # cartopy GeoAxes generated.
        plt.subplot(grid_size, grid_size, subplot)
        # Reshape cube data into a single array to allow for a single histogram.
        # Otherwise we plot xdim histograms stacked.
        member_data_1d = (member.data).flatten()
        plt.hist(member_data_1d, density=True, histtype=histtype, stacked=True)
        ax = plt.gca()
        ax.set_title(f"Member #{member.coord(stamp_coordinate).points[0]}")
        ax.set_xlim(vmin, vmax)

    # Overall figure title.
    fig.suptitle(title)

    fig.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())
    logging.info("Saved histogram postage stamp plot to %s", filename)
    plt.close(fig)


def _plot_and_save_postage_stamps_in_single_plot_histogram_series(
    cube: iris.cube.Cube,
    filename: str,
    title: str,
    stamp_coordinate: str,
    vmin: float,
    vmax: float,
    histtype: str = "step",
    **kwargs,
):
    fig, ax = plt.subplots(figsize=(10, 10), facecolor="w", edgecolor="k")
    ax.set_title(title)
    ax.set_xlim(vmin, vmax)
    ax.set_xlabel(f"{cube.name()} / {cube.units}")
    ax.set_ylabel("normalised probability density")
    # Loop over all slices along the stamp_coordinate
    for member in cube.slices_over(stamp_coordinate):
        # Flatten the member data to 1D
        member_data_1d = member.data.flatten()
        # Plot the histogram using plt.hist
        plt.hist(
            member_data_1d,
            density=True,
            histtype=histtype,
            stacked=True,
            label=f"Member #{member.coord(stamp_coordinate).points[0]}",
        )

    # Add a legend
    ax.legend()

    # Save the figure to a file
    plt.savefig(filename, bbox_inches="tight", dpi=_get_plot_resolution())

    # Close the figure
    plt.close(fig)


def _spatial_plot(
    method: Literal["contourf", "pcolormesh"],
    cube: iris.cube.Cube,
    filename: str | None,
    sequence_coordinate: str,
    stamp_coordinate: str,
):
    """Plot a spatial variable onto a map from a 2D, 3D, or 4D cube.

    A 2D spatial field can be plotted, but if the sequence_coordinate is present
    then a sequence of plots will be produced. Similarly if the stamp_coordinate
    is present then postage stamp plots will be produced.

    Parameters
    ----------
    method: "contourf" | "pcolormesh"
        The plotting method to use.
    cube: Cube
        Iris cube of the data to plot. It should have two spatial dimensions,
        such as lat and lon, and may also have a another two dimension to be
        plotted sequentially and/or as postage stamp plots.
    filename: str | None
        Name of the plot to write, used as a prefix for plot sequences. If None
        uses the recipe name.
    sequence_coordinate: str
        Coordinate about which to make a plot sequence. Defaults to ``"time"``.
        This coordinate must exist in the cube.
    stamp_coordinate: str
        Coordinate about which to plot postage stamp plots. Defaults to
        ``"realization"``.

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions.
    TypeError
        If the cube isn't a single cube.
    """
    recipe_title = get_recipe_metadata().get("title", "Untitled")

    # Ensure we have a name for the plot file.
    if filename is None:
        filename = slugify(recipe_title)

    # Ensure we've got a single cube.
    cube = _check_single_cube(cube)

    # Make postage stamp plots if stamp_coordinate exists and has more than a
    # single point.
    plotting_func = _plot_and_save_spatial_plot
    try:
        if cube.coord(stamp_coordinate).shape[0] > 1:
            plotting_func = _plot_and_save_postage_stamp_spatial_plot
    except iris.exceptions.CoordinateNotFoundError:
        pass

    # Must have a sequence coordinate.
    try:
        cube.coord(sequence_coordinate)
    except iris.exceptions.CoordinateNotFoundError as err:
        raise ValueError(f"Cube must have a {sequence_coordinate} coordinate.") from err

    # Create a plot for each value of the sequence coordinate.
    plot_index = []
    for cube_slice in cube.slices_over(sequence_coordinate):
        # Use sequence value so multiple sequences can merge.
        sequence_value = cube_slice.coord(sequence_coordinate).points[0]
        plot_filename = f"{filename.rsplit('.', 1)[0]}_{sequence_value}.png"
        coord = cube_slice.coord(sequence_coordinate)
        # Format the coordinate value in a unit appropriate way.
        title = f"{recipe_title}\n{coord.units.title(coord.points[0])}"
        # Do the actual plotting.
        plotting_func(
            cube_slice,
            filename=plot_filename,
            stamp_coordinate=stamp_coordinate,
            title=title,
            method=method,
        )
        plot_index.append(plot_filename)

    # Add list of plots to plot metadata.
    complete_plot_index = _append_to_plot_index(plot_index)

    # Make a page to display the plots.
    _make_plot_html_page(complete_plot_index)


####################
# Public functions #
####################


def spatial_contour_plot(
    cube: iris.cube.Cube,
    filename: str = None,
    sequence_coordinate: str = "time",
    stamp_coordinate: str = "realization",
    **kwargs,
) -> iris.cube.Cube:
    """Plot a spatial variable onto a map from a 2D, 3D, or 4D cube.

    A 2D spatial field can be plotted, but if the sequence_coordinate is present
    then a sequence of plots will be produced. Similarly if the stamp_coordinate
    is present then postage stamp plots will be produced.

    Parameters
    ----------
    cube: Cube
        Iris cube of the data to plot. It should have two spatial dimensions,
        such as lat and lon, and may also have a another two dimension to be
        plotted sequentially and/or as postage stamp plots.
    filename: str, optional
        Name of the plot to write, used as a prefix for plot sequences. Defaults
        to the recipe name.
    sequence_coordinate: str, optional
        Coordinate about which to make a plot sequence. Defaults to ``"time"``.
        This coordinate must exist in the cube.
    stamp_coordinate: str, optional
        Coordinate about which to plot postage stamp plots. Defaults to
        ``"realization"``.

    Returns
    -------
    Cube
        The original cube (so further operations can be applied).

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions.
    TypeError
        If the cube isn't a single cube.
    """
    _spatial_plot("contourf", cube, filename, sequence_coordinate, stamp_coordinate)
    return cube


def spatial_pcolormesh_plot(
    cube: iris.cube.Cube,
    filename: str = None,
    sequence_coordinate: str = "time",
    stamp_coordinate: str = "realization",
    **kwargs,
) -> iris.cube.Cube:
    """Plot a spatial variable onto a map from a 2D, 3D, or 4D cube.

    A 2D spatial field can be plotted, but if the sequence_coordinate is present
    then a sequence of plots will be produced. Similarly if the stamp_coordinate
    is present then postage stamp plots will be produced.

    This function is significantly faster than ``spatial_contour_plot``,
    especially at high resolutions, and should be preferred unless contiguous
    contour areas are important.

    Parameters
    ----------
    cube: Cube
        Iris cube of the data to plot. It should have two spatial dimensions,
        such as lat and lon, and may also have a another two dimension to be
        plotted sequentially and/or as postage stamp plots.
    filename: str, optional
        Name of the plot to write, used as a prefix for plot sequences. Defaults
        to the recipe name.
    sequence_coordinate: str, optional
        Coordinate about which to make a plot sequence. Defaults to ``"time"``.
        This coordinate must exist in the cube.
    stamp_coordinate: str, optional
        Coordinate about which to plot postage stamp plots. Defaults to
        ``"realization"``.

    Returns
    -------
    Cube
        The original cube (so further operations can be applied).

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions.
    TypeError
        If the cube isn't a single cube.
    """
    _spatial_plot("pcolormesh", cube, filename, sequence_coordinate, stamp_coordinate)
    return cube


# TODO: Expand function to handle ensemble data.
# line_coordinate: str, optional
#     Coordinate about which to plot multiple lines. Defaults to
#     ``"realization"``.
def plot_line_series(
    cube: iris.cube.Cube | iris.cube.CubeList,
    filename: str = None,
    series_coordinate: str = "time",
    # line_coordinate: str = "realization",
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Plot a line plot for the specified coordinate.

    The Cube or CubeList must be 1D.

    Parameters
    ----------
    iris.cube | iris.cube.CubeList
        Cube or CubeList of the data to plot. The individual cubes should have a single dimension.
        The cubes should cover the same phenomenon i.e. all cubes contain temperature data.
        We do not support different data such as temperature and humidity in the same CubeList for plotting.
    filename: str, optional
        Name of the plot to write, used as a prefix for plot sequences. Defaults
        to the recipe name.
    series_coordinate: str, optional
        Coordinate about which to make a series. Defaults to ``"time"``. This
        coordinate must exist in the cube.

    Returns
    -------
    iris.cube.Cube | iris.cube.CubeList
        The original Cube or CubeList (so further operations can be applied).
        plotted data.

    Raises
    ------
    ValueError
        If the cubes don't have the right dimensions.
    TypeError
        If the cube isn't a Cube or CubeList.
    """
    # Ensure we have a name for the plot file.
    title = get_recipe_metadata().get("title", "Untitled")
    if filename is None:
        filename = slugify(title)

    # Add file extension.
    plot_filename = f"{filename.rsplit('.', 1)[0]}.png"

    # Iterate over all cubes in cube or CubeList and plot.
    for cube_iter in iter_maybe(cube):
        # Check cube is right shape.
        cube_iter = _check_single_cube(cube_iter)
        try:
            coord = cube_iter.coord(series_coordinate)
        except iris.exceptions.CoordinateNotFoundError as err:
            raise ValueError(
                f"Cube must have a {series_coordinate} coordinate."
            ) from err
        if cube_iter.ndim > 1:
            raise ValueError("Cube must be 1D.")

    # Do the actual plotting.
    _plot_and_save_line_series(cube, coord, plot_filename, title)

    # Add list of plots to plot metadata.
    plot_index = _append_to_plot_index([plot_filename])

    # Make a page to display the plots.
    _make_plot_html_page(plot_index)

    return cube


def plot_vertical_line_series(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    filename: str = None,
    series_coordinate: str = "model_level_number",
    sequence_coordinate: str = "time",
    # line_coordinate: str = "realization",
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Plot a line plot against a type of vertical coordinate.

    The Cube or CubeList must be 1D.

    A 1D line plot with y-axis as pressure coordinate can be plotted, but if the sequence_coordinate is present
    then a sequence of plots will be produced.

    Parameters
    ----------
    iris.cube | iris.cube.CubeList
        Cube or CubeList of the data to plot. The individual cubes should have a single dimension.
        The cubes should cover the same phenomenon i.e. all cubes contain temperature data.
        We do not support different data such as temperature and humidity in the same CubeList for plotting.
    filename: str, optional
        Name of the plot to write, used as a prefix for plot sequences. Defaults
        to the recipe name.
    series_coordinate: str, optional
        Coordinate to plot on the y-axis. Can be ``pressure`` or
        ``model_level_number`` for UM, or ``full_levels`` or ``half_levels``
        for LFRic. Defaults to ``model_level_number``.
        This coordinate must exist in the cube.
    sequence_coordinate: str, optional
        Coordinate about which to make a plot sequence. Defaults to ``"time"``.
        This coordinate must exist in the cube.

    Returns
    -------
    iris.cube.Cube | iris.cube.CubeList
        The original Cube or CubeList (so further operations can be applied).
        Plotted data.

    Raises
    ------
    ValueError
        If the cubes doesn't have the right dimensions.
    TypeError
        If the cube isn't a Cube or CubeList.
    """
    # Ensure we have a name for the plot file.
    recipe_title = get_recipe_metadata().get("title", "Untitled")
    if filename is None:
        filename = slugify(recipe_title)

    # Initialise empty list to hold all data from all cubes in a CubeList
    all_data = []

    # Iterate over all cubes in cube or CubeList and plot.
    for cube_iter in iter_maybe(cubes):
        # Ensure we've got a single cube.
        cube_iter = _check_single_cube(cube_iter)

        # Test if series coordinate i.e. pressure level exist for any cube with cube.ndim >=1.
        try:
            coord = cube_iter.coord(series_coordinate)
        except iris.exceptions.CoordinateNotFoundError as err:
            raise ValueError(
                f"Cube must have a {series_coordinate} coordinate."
            ) from err

        try:
            if cube_iter.ndim > 1:
                cube_iter.coord(sequence_coordinate)
        except iris.exceptions.CoordinateNotFoundError as err:
            raise ValueError(
                f"Cube must have a {sequence_coordinate} coordinate or be 1D."
            ) from err

        # Append cube data to the list to calculate vmin and vmax across entire
        # cubelist.
        all_data.append(cube_iter.data)

    # Combine all data into a single NumPy array
    combined_data = np.concatenate(all_data)

    # Set the lower and upper limit for the x-axis to ensure all plots have same
    # range. This needs to read the whole cube over the range of the sequence
    # and if applicable postage stamp coordinate.
    vmin = np.floor(combined_data.min())
    vmax = np.ceil(combined_data.max())

    # Create a plot for each value of the sequence coordinate.
    # Allowing for multiple cubes in a CubeList to be plotted in the same plot for
    # similar sequence values. Passing a CubeList into the internal plotting function
    # for similar values of the sequence coordinate. cube_slice can be an iris.cube.Cube
    # or an iris.cube.CubeList.
    plot_index = []
    for cubes_slice in cubes.slices_over(sequence_coordinate):
        # Use sequence value so multiple sequences can merge.
        seq_coord = cubes_slice.coord(sequence_coordinate)
        sequence_value = seq_coord.points[0]
        plot_filename = f"{filename.rsplit('.', 1)[0]}_{sequence_value}.png"
        # Format the coordinate value in a unit appropriate way.
        title = f"{recipe_title}\n{seq_coord.units.title(sequence_value)}"
        # Do the actual plotting.
        _plot_and_save_vertical_line_series(
            cubes_slice,
            coord,
            plot_filename,
            series_coordinate,
            title=title,
            vmin=vmin,
            vmax=vmax,
        )
        plot_index.append(plot_filename)

    # Add list of plots to plot metadata.
    complete_plot_index = _append_to_plot_index(plot_index)

    # Make a page to display the plots.
    _make_plot_html_page(complete_plot_index)

    return cubes


def scatter_plot(
    cube_x: iris.cube.Cube | iris.cube.CubeList,
    cube_y: iris.cube.Cube | iris.cube.CubeList,
    filename: str = None,
    one_to_one: bool = True,
    **kwargs,
) -> iris.cube.CubeList:
    """Plot a scatter plot between two variables.

    Both cubes must be 1D.

    Parameters
    ----------
    cube_x: Cube | CubeList
        1 dimensional Cube of the data to plot on y-axis.
    cube_y: Cube | CubeList
        1 dimensional Cube of the data to plot on x-axis.
    filename: str, optional
        Filename of the plot to write.
    one_to_one: bool, optional
        If True a 1:1 line is plotted; if False it is not. Default is True.

    Returns
    -------
    cubes: CubeList
        CubeList of the original x and y cubes for further processing.

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions and cubes not the same
        size.
    TypeError
        If the cube isn't a single cube.

    Notes
    -----
    Scatter plots are used for determining if there is a relationship between
    two variables. Positive relations have a slope going from bottom left to top
    right; Negative relations have a slope going from top left to bottom right.

    A variant of the scatter plot is the quantile-quantile plot. This plot does
    not use all data points, but the selected quantiles of each variable
    instead. Quantile-quantile plots are valuable for comparing against
    observations and other models. Identical percentiles between the variables
    will lie on the one-to-one line implying the values correspond well to each
    other. Where there is a deviation from the one-to-one line a range of
    possibilities exist depending on how and where the data is shifted (e.g.,
    Wilks 2011 [Wilks2011]_).

    For distributions above the one-to-one line the distribution is left-skewed;
    below is right-skewed. A distinct break implies a bimodal distribution, and
    closer values/values further apart at the tails imply poor representation of
    the extremes.

    References
    ----------
    .. [Wilks2011] Wilks, D.S., (2011) "Statistical Methods in the Atmospheric
       Sciences" Third Edition, vol. 100, Academic Press, Oxford, UK, 676 pp.
    """
    # Iterate over all cubes in cube or CubeList and plot.
    for cube_iter in iter_maybe(cube_x):
        # Check cubes are correct shape.
        cube_iter = _check_single_cube(cube_iter)
        if cube_iter.ndim > 1:
            raise ValueError("cube_x must be 1D.")

    # Iterate over all cubes in cube or CubeList and plot.
    for cube_iter in iter_maybe(cube_y):
        # Check cubes are correct shape.
        cube_iter = _check_single_cube(cube_iter)
        if cube_iter.ndim > 1:
            raise ValueError("cube_y must be 1D.")

    # Ensure we have a name for the plot file.
    title = get_recipe_metadata().get("title", "Untitled")
    if filename is None:
        filename = slugify(title)

    # Add file extension.
    plot_filename = f"{filename.rsplit('.', 1)[0]}.png"

    # Do the actual plotting.
    _plot_and_save_scatter_plot(cube_x, cube_y, plot_filename, title, one_to_one)

    # Add list of plots to plot metadata.
    plot_index = _append_to_plot_index([plot_filename])

    # Make a page to display the plots.
    _make_plot_html_page(plot_index)

    return iris.cube.CubeList([cube_x, cube_y])


def plot_histogram_series(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    filename: str = None,
    sequence_coordinate: str = "time",
    stamp_coordinate: str = "realization",
    single_plot: bool = False,
    histtype: str = "step",
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Plot a histogram plot for each vertical level provided.

    A histogram plot can be plotted, but if the sequence_coordinate (i.e. time)
    is present then a sequence of plots will be produced using the time slider
    functionality to scroll through histograms against time. If a
    stamp_coordinate is present then postage stamp plots will be produced. If
    stamp_coordinate and single_plot is True, all postage stamp plots will be
    plotted in a single plot instead of separate postage stamp plots.

    Parameters
    ----------
    cubes: Cube | iris.cube.CubeList
        Iris cube or CubeList of the data to plot. It should have a single dimension other
        than the stamp coordinate.
        The cubes should cover the same phenomenon i.e. all cubes contain temperature data.
        We do not support different data such as temperature and humidity in the same CubeList for plotting.
    filename: str, optional
        Name of the plot to write, used as a prefix for plot sequences. Defaults
        to the recipe name.
    sequence_coordinate: str, optional
        Coordinate about which to make a plot sequence. Defaults to ``"time"``.
        This coordinate must exist in the cube and will be used for the time
        slider.
    stamp_coordinate: str, optional
        Coordinate about which to plot postage stamp plots. Defaults to
        ``"realization"``.
    single_plot: bool, optional
        If True, all postage stamp plots will be plotted in a single plot. If
        False, each postage stamp plot will be plotted separately. Is only valid
        if stamp_coordinate exists and has more than a single point.
    histtype: str, optional
        The type of histogram to plot. Options are "step" for a line histogram
        or "barstacked", "stepfilled". "Step" is the default option, but can be
        changed in the rose-suite.conf configuration.

    Returns
    -------
    iris.cube.Cube | iris.cube.CubeList
        The original Cube or CubeList (so further operations can be applied).
        Plotted data.

    Raises
    ------
    ValueError
        If the cube doesn't have the right dimensions.
    TypeError
        If the cube isn't a Cube or CubeList.
    """
    recipe_title = get_recipe_metadata().get("title", "Untitled")

    # Ensure we have a name for the plot file.
    if filename is None:
        filename = slugify(recipe_title)

    # Internal plotting function.
    plotting_func = _plot_and_save_histogram_series

    # Initialise empty list to hold all data from all cubes in a CubeList
    all_data = []

    # Iterate over all cubes in cube or CubeList and plot.
    for cube_iter in iter_maybe(cubes):
        # Ensure we've got a single cube.
        cube_iter = _check_single_cube(cube_iter)

        # Make postage stamp plots if stamp_coordinate exists and has more than a
        # single point. If single_plot is True, all postage stamp plots will be
        # plotted in a single plot instead of separate postage stamp plots.
        if (
            stamp_coordinate in [c.name() for c in cube_iter.coords()]
            and cube_iter.coord(stamp_coordinate).shape[0] > 1
        ):
            if single_plot:
                plotting_func = (
                    _plot_and_save_postage_stamps_in_single_plot_histogram_series
                )
            else:
                plotting_func = _plot_and_save_postage_stamp_histogram_series

        # If several histograms are plotted with time as sequence_coordinate for the
        # time slider option.
        try:
            cube_iter.coord(sequence_coordinate)
        except iris.exceptions.CoordinateNotFoundError as err:
            raise ValueError(
                f"Cube must have a {sequence_coordinate} coordinate."
            ) from err

        # append cube data to the list to calculate vmin and vmax across entire cubelist
        all_data.append(cube_iter.data)

    # Combine all data into a single NumPy array
    combined_data = np.concatenate(all_data)

    # Set the lower and upper limit for the colorbar to ensure all plots have
    # same range. This needs to read the whole cube over the range of the
    # sequence and if applicable postage stamp coordinate.
    vmin = np.floor((combined_data.min()))
    vmax = np.ceil((combined_data.max()))

    # Create a plot for each value of the sequence coordinate.
    # Allowing for multiple cubes in a CubeList to be plotted in the same plot for
    # similar sequence values. Passing a CubeList into the internal plotting function
    # for similar values of the sequence coordinate. cube_slice can be an iris.cube.Cube
    # or an iris.cube.CubeList.
    plot_index = []
    for cube_slice in cubes.slices_over(sequence_coordinate):
        # Use sequence value so multiple sequences can merge.
        sequence_value = cube_slice.coord(sequence_coordinate).points[0]
        plot_filename = f"{filename.rsplit('.', 1)[0]}_{sequence_value}.png"
        coord = cube_slice.coord(sequence_coordinate)
        # Format the coordinate value in a unit appropriate way.
        title = f"{recipe_title}\n{coord.units.title(coord.points[0])}"
        # Do the actual plotting.
        plotting_func(
            cube_slice,
            plot_filename,
            stamp_coordinate=stamp_coordinate,
            title=title,
            vmin=vmin,
            vmax=vmax,
            histtype=histtype,
        )
        plot_index.append(plot_filename)

    # Add list of plots to plot metadata.
    complete_plot_index = _append_to_plot_index(plot_index)

    # Make a page to display the plots.
    _make_plot_html_page(complete_plot_index)

    return cubes



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for finish_website workflow utility."""

import json
import logging

from CSET._workflow_utils import finish_website


def test_write_workflow_status(monkeypatch, tmp_path):
    """Workflow  finish status gets written to status file."""
    web_dir = tmp_path / "web"
    web_dir.mkdir()
    monkeypatch.setenv("CYLC_WORKFLOW_SHARE_DIR", str(tmp_path))
    finish_website.update_workflow_status()
    with open(web_dir / "status.html", "rt", encoding="UTF-8") as fp:
        assert fp.read() == "<p>Finished</p>\n"


def test_construct_index(monkeypatch, tmp_path):
    """Test putting the index together."""
    monkeypatch.setenv("CYLC_WORKFLOW_SHARE_DIR", str(tmp_path))
    plots_dir = tmp_path / "web/plots"
    plots_dir.mkdir(parents=True)

    # Plot directories.
    plot1 = plots_dir / "p1/meta.json"
    plot1.parent.mkdir()
    plot1.write_text('{"category": "Category", "title": "P1"}')
    plot2 = plots_dir / "p2/meta.json"
    plot2.parent.mkdir()
    plot2.write_text('{"category": "Category", "title": "P2"}')

    # Non-plot directory also present.
    static_resource = plots_dir / "static/style.css"
    static_resource.parent.mkdir()
    static_resource.touch()

    # Construct index.
    finish_website.construct_index()

    # Check index.
    index_file = plots_dir / "index.json"
    assert index_file.is_file()
    with open(index_file, "rt", encoding="UTF-8") as fp:
        index = json.load(fp)
    expected = {"Category": {"p1": "P1", "p2": "P2"}}
    assert index == expected


def test_construct_index_invalid(monkeypatch, tmp_path, caplog):
    """Test constructing index when metadata is invalid."""
    monkeypatch.setenv("CYLC_WORKFLOW_SHARE_DIR", str(tmp_path))
    plots_dir = tmp_path / "web/plots"
    plots_dir.mkdir(parents=True)

    # Plot directories.
    plot = plots_dir / "p1/meta.json"
    plot.parent.mkdir()
    plot.write_text('"Not JSON!"')

    # Construct index.
    finish_website.construct_index()

    # Check log message.
    _, level, message = caplog.record_tuples[0]
    assert level == logging.ERROR
    assert "p1/meta.json is invalid, skipping." in message

    index_file = plots_dir / "index.json"
    assert index_file.is_file()
    with open(index_file, "rt", encoding="UTF-8") as fp:
        index = json.load(fp)
    expected = {}
    assert index == expected


def test_entrypoint(monkeypatch):
    """Test running the finish_website module."""
    # Count the number of times the other functions are run, to ensure they
    # are both run.
    counter = 0

    def increment_counter():
        nonlocal counter
        counter += 1

    monkeypatch.setattr(finish_website, "construct_index", increment_counter)
    monkeypatch.setattr(finish_website, "update_workflow_status", increment_counter)

    # Just check that it runs all the needed subfunctions.
    finish_website.run()
    assert counter == 2



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test mesoscale operators."""

import numpy as np
from scipy.ndimage import gaussian_filter, uniform_filter

import CSET.operators.mesoscale as mesoscale
from CSET.operators._utils import get_cube_yxcoordname


def test_spatial_perturbation_field_gaussian(cube):
    """Test smoothing a cube with a Gaussian filter."""
    calculated = cube.copy()
    coords = [coord.name() for coord in cube.coords()]
    axes = (
        coords.index(get_cube_yxcoordname(cube)[0]),
        coords.index(get_cube_yxcoordname(cube)[1]),
    )
    calculated.data -= gaussian_filter(cube.data, 40, axes=axes)
    assert np.allclose(
        calculated.data,
        mesoscale.spatial_perturbation_field(cube).data,
        rtol=1e-06,
        atol=1e-02,
    )

    calculated_2 = cube.copy()
    calculated_2.data -= gaussian_filter(cube.data, 100, axes=axes)
    assert np.allclose(
        calculated_2.data,
        mesoscale.spatial_perturbation_field(cube, filter_scale=100).data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_spatial_perturbation_field_uniform(cube):
    """Test smoothing a cube with a uniform filter."""
    calculated = cube.copy()
    coords = [coord.name() for coord in cube.coords()]
    axes = (
        coords.index(get_cube_yxcoordname(cube)[0]),
        coords.index(get_cube_yxcoordname(cube)[1]),
    )
    calculated.data -= uniform_filter(cube.data, 40, axes=axes)
    assert np.allclose(
        calculated.data,
        mesoscale.spatial_perturbation_field(cube, apply_gaussian_filter=False).data,
        rtol=1e-06,
        atol=1e-02,
    )

    calculated_2 = cube.copy()
    calculated_2.data -= uniform_filter(cube.data, 100, axes=axes)
    assert np.allclose(
        calculated_2.data,
        mesoscale.spatial_perturbation_field(
            cube, apply_gaussian_filter=False, filter_scale=100
        ).data,
        rtol=1e-06,
        atol=1e-02,
    )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Recipe tests."""

from pathlib import Path

import pytest

from CSET import recipes


def test_recipe_files_in_tree():
    """Get recipes in directory containing sub-directories."""
    files = recipes._recipe_files_in_tree(input_dir=Path("tests"))
    assert Path("tests/test_data/noop_recipe.yaml") in files


def test_recipe_files_in_tree_from_package():
    """Get a recipe from inside the CSET package."""
    files = recipes._recipe_files_in_tree()
    assert any("CAPE_ratio_plot.yaml" == path.name for path in files)


def test_unpack(tmp_path: Path):
    """Unpack recipes."""
    recipes.unpack_recipe(tmp_path, "CAPE_ratio_plot.yaml")
    assert (tmp_path / "CAPE_ratio_plot.yaml").is_file()
    # Unpack everything and check a warning is produced when files collide.
    with pytest.warns():
        recipes.unpack_recipe(tmp_path, "CAPE_ratio_plot.yaml")


def test_unpack_recipes_exception_collision(tmp_path: Path):
    """Output path already exists and is not a directory."""
    file_path = tmp_path / "regular_file"
    file_path.touch()
    with pytest.raises(FileExistsError):
        recipes.unpack_recipe(file_path, "CAPE_ratio_plot.yaml")


def test_unpack_recipes_exception_permission():
    """Insufficient permission to create output directory.

    Will fail if tests are run as root, but no one should do that, right?
    """
    with pytest.raises(OSError):
        recipes.unpack_recipe(Path("/usr/bin/cset"), "extract_instant_air_temp.yaml")


def test_get_recipe_file():
    """Get a recipe file from a specific location."""
    file = recipes._get_recipe_file("noop_recipe.yaml", Path("tests/test_data"))
    assert file.is_file()


def test_get_recipe_file_missing():
    """Exception raised when recipe file not in location."""
    with pytest.raises(FileNotFoundError):
        recipes._get_recipe_file("non-existent", Path("tests/test_data"))


def test_get_recipe_file_in_package():
    """Get a recipe file from a the default location inside the package."""
    file = recipes._get_recipe_file("CAPE_ratio_plot.yaml")
    assert file.is_file()


def test_list_available_recipes(capsys):
    """Display available recipes."""
    recipes.list_available_recipes()
    # Read stdout and stderr.
    captured_output = capsys.readouterr()
    # Check start.
    assert captured_output.out.startswith("Available recipes:\n")
    # Check has some recipes.
    assert len(captured_output.out.splitlines()) > 3


def test_detail_recipe(capsys):
    """Show detail of a recipe."""
    recipes.detail_recipe("CAPE_ratio_plot.yaml")
    # Read stdout and stderr.
    captured_output = capsys.readouterr()
    assert captured_output.out.startswith("\n\tCAPE_ratio_plot.yaml\n")
